###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# ruff: noqa: E501
"""Update node types after `core.` prefix was added to entry point names.

Revision ID: 34a831f4286d
Revises: 535039300e4a
Create Date: 2021-08-11 18:25:48.706298

"""

from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = '34a831f4286d'
down_revision = '535039300e4a'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()
    statement = text(
        """
        UPDATE db_dbnode SET node_type = 'data.core.array.ArrayData.' WHERE node_type = 'data.array.ArrayData.';
        UPDATE db_dbnode SET node_type = 'data.core.array.bands.BandsData.' WHERE node_type = 'data.array.bands.BandsData.';
        UPDATE db_dbnode SET node_type = 'data.core.array.kpoints.KpointsData.' WHERE node_type = 'data.array.kpoints.KpointsData.';
        UPDATE db_dbnode SET node_type = 'data.core.array.projection.ProjectionData.' WHERE node_type = 'data.array.projection.ProjectionData.';
        UPDATE db_dbnode SET node_type = 'data.core.array.trajectory.TrajectoryData.' WHERE node_type = 'data.array.trajectory.TrajectoryData.';
        UPDATE db_dbnode SET node_type = 'data.core.array.xy.XyData.' WHERE node_type = 'data.array.xy.XyData.';
        UPDATE db_dbnode SET node_type = 'data.core.base.BaseData.' WHERE node_type = 'data.base.BaseData.';
        UPDATE db_dbnode SET node_type = 'data.core.bool.Bool.' WHERE node_type = 'data.bool.Bool.';
        UPDATE db_dbnode SET node_type = 'data.core.cif.CifData.' WHERE node_type = 'data.cif.CifData.';
        UPDATE db_dbnode SET node_type = 'data.core.code.Code.' WHERE node_type = 'data.code.Code.';
        UPDATE db_dbnode SET node_type = 'data.core.dict.Dict.' WHERE node_type = 'data.dict.Dict.';
        UPDATE db_dbnode SET node_type = 'data.core.float.Float.' WHERE node_type = 'data.float.Float.';
        UPDATE db_dbnode SET node_type = 'data.core.folder.FolderData.' WHERE node_type = 'data.folder.FolderData.';
        UPDATE db_dbnode SET node_type = 'data.core.int.Int.' WHERE node_type = 'data.int.Int.';
        UPDATE db_dbnode SET node_type = 'data.core.list.List.' WHERE node_type = 'data.list.List.';
        UPDATE db_dbnode SET node_type = 'data.core.numeric.NumericData.' WHERE node_type = 'data.numeric.NumericData.';
        UPDATE db_dbnode SET node_type = 'data.core.orbital.OrbitalData.' WHERE node_type = 'data.orbital.OrbitalData.';
        UPDATE db_dbnode SET node_type = 'data.core.remote.RemoteData.' WHERE node_type = 'data.remote.RemoteData.';
        UPDATE db_dbnode SET node_type = 'data.core.remote.stash.RemoteStashData.' WHERE node_type = 'data.remote.stash.RemoteStashData.';
        UPDATE db_dbnode SET node_type = 'data.core.remote.stash.folder.RemoteStashFolderData.' WHERE node_type = 'data.remote.stash.folder.RemoteStashFolderData.';
        UPDATE db_dbnode SET node_type = 'data.core.singlefile.SinglefileData.' WHERE node_type = 'data.singlefile.SinglefileData.';
        UPDATE db_dbnode SET node_type = 'data.core.str.Str.' WHERE node_type = 'data.str.Str.';
        UPDATE db_dbnode SET node_type = 'data.core.structure.StructureData.' WHERE node_type = 'data.structure.StructureData.';
        UPDATE db_dbnode SET node_type = 'data.core.upf.UpfData.' WHERE node_type = 'data.upf.UpfData.';
        UPDATE db_dbcomputer SET scheduler_type = 'core.direct' WHERE scheduler_type = 'direct';
        UPDATE db_dbcomputer SET scheduler_type = 'core.lsf' WHERE scheduler_type = 'lsf';
        UPDATE db_dbcomputer SET scheduler_type = 'core.pbspro' WHERE scheduler_type = 'pbspro';
        UPDATE db_dbcomputer SET scheduler_type = 'core.sge' WHERE scheduler_type = 'sge';
        UPDATE db_dbcomputer SET scheduler_type = 'core.slurm' WHERE scheduler_type = 'slurm';
        UPDATE db_dbcomputer SET scheduler_type = 'core.torque' WHERE scheduler_type = 'torque';
        UPDATE db_dbcomputer SET transport_type = 'core.local' WHERE transport_type = 'local';
        UPDATE db_dbcomputer SET transport_type = 'core.ssh' WHERE transport_type = 'ssh';
        UPDATE db_dbnode SET process_type = 'aiida.calculations:core.arithmetic.add' WHERE process_type = 'aiida.calculations:arithmetic.add';
        UPDATE db_dbnode SET process_type = 'aiida.calculations:core.templatereplacer' WHERE process_type = 'aiida.calculations:templatereplacer';
        UPDATE db_dbnode SET process_type = 'aiida.workflows:core.arithmetic.add_multiply' WHERE process_type = 'aiida.workflows:arithmetic.add_multiply';
        UPDATE db_dbnode SET process_type = 'aiida.workflows:core.arithmetic.multiply_add' WHERE process_type = 'aiida.workflows:arithmetic.multiply_add';
        UPDATE db_dbnode SET attributes = jsonb_set(attributes, '{"parser_name"}', '"core.arithmetic.add"') WHERE attributes->>'parser_name' = 'arithmetic.add';
        UPDATE db_dbnode SET attributes = jsonb_set(attributes, '{"parser_name"}', '"core.templatereplacer.doubler"') WHERE attributes->>'parser_name' = 'templatereplacer.doubler';
        """
    )
    conn.execute(statement)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of 34a831f4286d.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of the ``SqliteTempBackend`` backend."""

from __future__ import annotations

import functools
import hashlib
import os
import shutil
from contextlib import contextmanager, nullcontext
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, BinaryIO, Iterator, Sequence

from pydantic import BaseModel, Field
from sqlalchemy import column, insert, update
from sqlalchemy.orm import Session

from aiida.common.exceptions import ClosedStorage, IntegrityError
from aiida.manage.configuration import Profile
from aiida.orm.entities import EntityTypes
from aiida.orm.implementation import BackendEntity, StorageBackend
from aiida.repository.backend.sandbox import SandboxRepositoryBackend
from aiida.storage.sqlite_zip import models, orm
from aiida.storage.sqlite_zip.migrator import get_schema_version_head
from aiida.storage.sqlite_zip.utils import create_sqla_engine

__all__ = ('SqliteTempBackend',)


class SqliteTempBackend(StorageBackend):
    """A temporary backend, using an in-memory sqlite database.

    This backend is intended for testing and demonstration purposes.
    Whenever it is instantiated, it creates a fresh storage backend,
    and destroys it when it is garbage collected.
    """

    class Model(BaseModel, defer_build=True):
        filepath: str = Field(
            title='Temporary directory',
            description='Temporary directory in which to store data for this backend.',
            default_factory=mkdtemp,
        )

    cli_exposed = False
    """Ensure this plugin is not exposed in ``verdi profile setup``."""

    @staticmethod
    def create_profile(
        name: str = 'temp',
        default_user_email='user@email.com',
        options: dict | None = None,
        debug: bool = False,
        filepath: str | Path | None = None,
    ) -> Profile:
        """Create a new profile instance for this backend, from the path to the zip file."""
        return Profile(
            name,
            {
                'default_user_email': default_user_email,
                'storage': {
                    'backend': 'core.sqlite_temp',
                    'config': {
                        'filepath': filepath,
                        'debug': debug,
                    },
                },
                'process_control': {'backend': None, 'config': {}},
                'options': options or {},
            },
        )

    @classmethod
    def version_head(cls) -> str:
        return get_schema_version_head()

    @classmethod
    def version_profile(cls, profile: Profile) -> str | None:
        return get_schema_version_head()

    @classmethod
    def initialise(cls, profile: 'Profile', reset: bool = False) -> bool:
        return False

    @classmethod
    def migrate(cls, profile: Profile):
        pass

    def __init__(self, profile: Profile):
        super().__init__(profile)
        self._session: Session | None = None
        self._repo: SandboxShaRepositoryBackend = SandboxShaRepositoryBackend(profile.storage_config['filepath'])
        self._globals: dict[str, tuple[Any, str | None]] = {}
        self._closed = False
        self.get_session()  # load the database on initialization

    def __str__(self) -> str:
        state = 'closed' if self.is_closed else 'open'
        return f'SqliteTemp storage [{state}], sandbox: {self._repo}'

    @property
    def is_closed(self) -> bool:
        return self._closed

    def close(self):
        if self._session:
            self._session.close()
        if self._repo:
            self._repo.erase()
        self._session = None
        self._repo = None
        self._globals = {}
        self._closed = True

    def get_global_variable(self, key: str):
        return self._globals[key][0]

    def set_global_variable(self, key: str, value, description: str | None = None, overwrite=True) -> None:
        if not overwrite and key in self._globals:
            raise ValueError(f'global variable {key} already exists')
        self._globals[key] = (value, description)

    def get_session(self) -> Session:
        """Return an SQLAlchemy session."""
        if self._closed:
            raise ClosedStorage(str(self))
        if self._session is None:
            engine = create_sqla_engine(':memory:', echo=self.profile.storage_config.get('debug', False))
            models.SqliteBase.metadata.create_all(engine)
            self._session = Session(engine, future=True)
            self._session.add(models.DbUser(email=self.profile.default_user_email or 'user@email.com'))
            self._session.commit()
        return self._session

    def get_repository(self) -> SandboxShaRepositoryBackend:
        if self._closed:
            raise ClosedStorage(str(self))
        return self._repo

    @property
    def in_transaction(self) -> bool:
        return self.get_session().in_nested_transaction()

    @contextmanager
    def transaction(self) -> Iterator[Session]:
        """Open a transaction to be used as a context manager.

        If there is an exception within the context then the changes will be rolled back and the state will be as before
        entering. Transactions can be nested.
        """
        session = self.get_session()
        if session.in_transaction():
            with session.begin_nested():
                yield session
            session.commit()
        else:
            with session.begin():
                with session.begin_nested():
                    yield session

    def _clear(self) -> None:
        raise NotImplementedError

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        pass

    def query(self) -> orm.SqliteQueryBuilder:
        return orm.SqliteQueryBuilder(self)

    def get_backend_entity(self, model) -> BackendEntity:
        """Return the backend entity that corresponds to the given Model instance."""
        return orm.get_backend_entity(model, self)

    @functools.cached_property
    def authinfos(self):
        return orm.SqliteAuthInfoCollection(self)

    @functools.cached_property
    def comments(self):
        return orm.SqliteCommentCollection(self)

    @functools.cached_property
    def computers(self):
        return orm.SqliteComputerCollection(self)

    @functools.cached_property
    def groups(self):
        return orm.SqliteGroupCollection(self)

    @functools.cached_property
    def logs(self):
        return orm.SqliteLogCollection(self)

    @functools.cached_property
    def nodes(self):
        return orm.SqliteNodeCollection(self)

    @functools.cached_property
    def users(self):
        return orm.SqliteUserCollection(self)

    def get_info(self, detailed: bool = False) -> dict:
        results = super().get_info(detailed=detailed)
        # results['repository'] = self.get_repository().get_info(detailed)
        return results

    @staticmethod
    @functools.lru_cache(maxsize=18)
    def _get_mapper_from_entity(entity_type: EntityTypes, with_pk: bool):
        """Return the Sqlalchemy mapper and fields corresponding to the given entity.

        :param with_pk: if True, the fields returned will include the primary key
        """
        from sqlalchemy import inspect

        from aiida.storage.sqlite_zip.models import (
            DbAuthInfo,
            DbComment,
            DbComputer,
            DbGroup,
            DbGroupNodes,
            DbLink,
            DbLog,
            DbNode,
            DbUser,
        )

        model = {
            EntityTypes.AUTHINFO: DbAuthInfo,
            EntityTypes.COMMENT: DbComment,
            EntityTypes.COMPUTER: DbComputer,
            EntityTypes.GROUP: DbGroup,
            EntityTypes.LOG: DbLog,
            EntityTypes.NODE: DbNode,
            EntityTypes.USER: DbUser,
            EntityTypes.LINK: DbLink,
            EntityTypes.GROUP_NODE: DbGroupNodes,
        }[entity_type]
        mapper = inspect(model).mapper
        keys = {key for key, col in mapper.c.items() if with_pk or col not in mapper.primary_key}
        return mapper, keys

    def bulk_insert(self, entity_type: EntityTypes, rows: list[dict], allow_defaults: bool = False) -> list[int]:
        mapper, keys = self._get_mapper_from_entity(entity_type, False)
        if not rows:
            return []
        if entity_type in (EntityTypes.COMPUTER, EntityTypes.LOG, EntityTypes.AUTHINFO):
            for row in rows:
                row['_metadata'] = row.pop('metadata')
        if allow_defaults:
            for row in rows:
                if not keys.issuperset(row):
                    raise IntegrityError(f'Incorrect fields given for {entity_type}: {set(row)} not subset of {keys}')
        else:
            for row in rows:
                if set(row) != keys:
                    raise IntegrityError(f'Incorrect fields given for {entity_type}: {set(row)} != {keys}')
        session = self.get_session()
        with nullcontext() if self.in_transaction else self.transaction():
            result = session.execute(insert(mapper).returning(mapper, column('id')), rows).fetchall()
        return [row.id for row in result]

    def bulk_update(self, entity_type: EntityTypes, rows: list[dict]) -> None:
        mapper, keys = self._get_mapper_from_entity(entity_type, True)
        if not rows:
            return None
        for row in rows:
            if 'id' not in row:
                raise IntegrityError(f"'id' field not given for {entity_type}: {set(row)}")
            if not keys.issuperset(row):
                raise IntegrityError(f'Incorrect fields given for {entity_type}: {set(row)} not subset of {keys}')
        session = self.get_session()
        with nullcontext() if self.in_transaction else self.transaction():
            session.execute(update(mapper), rows)

    def delete(self) -> None:
        """Delete the storage and all the data."""
        self._repo.erase()

    def delete_nodes_and_connections(self, pks_to_delete: Sequence[int]):
        raise NotImplementedError


class SandboxShaRepositoryBackend(SandboxRepositoryBackend):
    """A sandbox repository backend that uses the sha256 of the file as the key.

    This allows for compatibility with the archive format (i.e. `SqliteZipBackend`).
    Which allows for temporary profiles to be exported and imported.
    """

    @property
    def key_format(self) -> str | None:
        return 'sha256'

    def get_object_hash(self, key: str) -> str:
        return key

    def _put_object_from_filelike(self, handle: BinaryIO) -> str:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :return: the generated fully qualified identifier for the object within the repository.
        :raises TypeError: if the handle is not a byte stream.
        """
        # we first compute the hash of the file contents
        hsh = hashlib.sha256()
        position = handle.tell()
        while True:
            buf = handle.read(1024 * 1024)
            if not buf:
                break
            hsh.update(buf)
        key = hsh.hexdigest()

        filepath = os.path.join(self.sandbox.abspath, key)
        if not os.path.exists(filepath):
            # if a file with this hash does not already exist
            # then we reset the file pointer and copy the contents
            handle.seek(position)
            with open(filepath, 'wb') as target:
                shutil.copyfileobj(handle, target)

        return key

    def get_info(self, detailed: bool = False, **kwargs) -> dict:
        return {'objects': {'count': len(list(self.list_objects()))}}

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        pass



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Delete trajectory symbols array from the repository and the reference in the attributes.

Note, this is similar to the sqlalchemy migration ce56d84bcc35

Revision ID: django_0027
Revises: django_0026

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.expression import delete

from aiida.storage.psql_dos.backend import get_filepath_container
from aiida.storage.psql_dos.migrations.utils import utils

revision = 'django_0027'
down_revision = 'django_0026'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    connection = op.get_bind()
    repo_path = get_filepath_container(op.get_context().opts['aiida_profile']).parent

    node_tbl = sa.table(
        'db_dbnode',
        sa.column('id', sa.Integer),
        sa.column('uuid', postgresql.UUID),
        sa.column('type', sa.String),
        # sa.column('attributes', JSONB),
    )

    nodes = connection.execute(
        sa.select(node_tbl.c.id, node_tbl.c.uuid).where(
            node_tbl.c.type == op.inline_literal('node.data.array.trajectory.TrajectoryData.')
        )
    ).all()

    attr_tbl = sa.table('db_dbattribute', sa.column('key'))

    for pk, uuid in nodes:
        connection.execute(delete(attr_tbl).where(sa.and_(node_tbl.c.id == pk, attr_tbl.c.key == 'array|symbols')))
        connection.execute(
            delete(attr_tbl).where(sa.and_(node_tbl.c.id == pk, attr_tbl.c.key.startswith('array|symbols.')))
        )
        utils.delete_numpy_array_from_repository(repo_path, uuid, 'symbols')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0027.')



"""Utilites for RabbitMQ."""

from . import defaults

__all__ = ('get_rmq_url', 'get_launch_queue_name', 'get_message_exchange_name', 'get_task_exchange_name')


def get_rmq_url(protocol=None, username=None, password=None, host=None, port=None, virtual_host=None, **kwargs):
    """Return the URL to connect to RabbitMQ.

    .. note::

        The default of the ``host`` is set to ``127.0.0.1`` instead of ``localhost`` because on some computers localhost
        resolves first to IPv6 with address ::1 and if RMQ is not running on IPv6 one gets an annoying warning. For more
        info see: https://github.com/aiidateam/aiida-core/issues/1142

    :param protocol: the protocol to use, `amqp` or `amqps`.
    :param username: the username for authentication.
    :param password: the password for authentication.
    :param host: the hostname of the RabbitMQ server.
    :param port: the port of the RabbitMQ server.
    :param virtual_host: the virtual host to connect to.
    :param kwargs: remaining keyword arguments that will be encoded as query parameters.
    :returns: the connection URL string.
    """
    from urllib.parse import urlencode, urlunparse

    if 'heartbeat' not in kwargs:
        kwargs['heartbeat'] = defaults.BROKER_DEFAULTS.heartbeat

    scheme = protocol or defaults.BROKER_DEFAULTS.protocol
    netloc = '{username}:{password}@{host}:{port}'.format(
        username=username or defaults.BROKER_DEFAULTS.username,
        password=password or defaults.BROKER_DEFAULTS.password,
        host=host or defaults.BROKER_DEFAULTS.host,
        port=port or defaults.BROKER_DEFAULTS.port,
    )
    path = virtual_host or defaults.BROKER_DEFAULTS.virtual_host
    parameters = ''
    query = urlencode(kwargs)
    fragment = ''

    # The virtual host is optional but if it is specified it needs to start with a forward slash. If the virtual host
    # itself contains forward slashes, they need to be encoded.
    if path and not path.startswith('/'):
        path = f'/{path}'

    return urlunparse((scheme, netloc, path, parameters, query, fragment))


def get_launch_queue_name(prefix=None):
    """Return the launch queue name with an optional prefix.

    :returns: launch queue name
    """
    if prefix is not None:
        return f'{prefix}.{defaults.LAUNCH_QUEUE}'

    return defaults.LAUNCH_QUEUE


def get_message_exchange_name(prefix):
    """Return the message exchange name for a given prefix.

    :returns: message exchange name
    """
    return f'{prefix}.{defaults.MESSAGE_EXCHANGE}'


def get_task_exchange_name(prefix):
    """Return the task exchange name for a given prefix.

    :returns: task exchange name
    """
    return f'{prefix}.{defaults.TASK_EXCHANGE}'



"""Module with utilities."""

from collections.abc import Mapping

from aiida.orm import Node


def prune_mapping(value):
    """Prune a nested mapping from all mappings that are completely empty.

    .. note:: A nested mapping that is completely empty means it contains at most other empty mappings. Other null
        values, such as `None` or empty lists, should not be pruned.

    :param value: A nested mapping of port values.
    :return: The same mapping but without any nested namespace that is completely empty.
    """
    if isinstance(value, Mapping) and not isinstance(value, Node):  # type: ignore[unreachable]
        result = {}
        for key, sub_value in value.items():
            pruned = prune_mapping(sub_value)
            # If `pruned` is an "empty'ish" mapping and not an instance of `Node`, skip it, otherwise keep it.
            if not (
                isinstance(pruned, Mapping) and not pruned and not isinstance(pruned, Node)  # type: ignore[unreachable]
            ):
                result[key] = pruned
        return result

    return value



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Define warnings that can be thrown by AiiDA."""

import os
import warnings


class AiidaDeprecationWarning(Warning):
    """Class for AiiDA deprecations.

    It does *not* inherit, on purpose, from `DeprecationWarning` as
    this would be filtered out by default.
    Enabled by default, you can disable it by running in the shell::

      verdi config warnings.showdeprecations False
    """


class AiidaEntryPointWarning(Warning):
    """Class for warnings concerning AiiDA entry points."""


class AiidaTestWarning(Warning):
    """Class for warnings concerning the AiiDA testing infrastructure."""


def warn_deprecation(message: str, version: int, stacklevel=2) -> None:
    """Warn about a deprecation for a future aiida-core version.

    Warnings are emitted if the ``warnings.showdeprecations`` config option is set to ``True``. Its value can be
    overwritten with the ``AIIDA_WARN_v{version}`` environment variable. The exact value for the environment variable is
    irrelevant. Any value will mean the variable is enabled and warnings will be emitted.

    :param message: the message to be printed
    :param version: the major version number of the future version
    :param stacklevel: the stack level at which the warning is issued
    """
    from aiida.manage.configuration import get_config_option

    from_config = get_config_option('warnings.showdeprecations')
    from_environment = os.environ.get(f'AIIDA_WARN_v{version}')

    if from_config or from_environment:
        message = f'{message} (this will be removed in v{version})'
        warnings.warn(message, AiidaDeprecationWarning, stacklevel=stacklevel)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Schema validation and migration utilities.

This code interacts directly with the database, outside of the ORM,
taking a `Profile` as input for the connection configuration.

.. important:: This code should only be accessed via the storage backend class, not directly!
"""

from __future__ import annotations

import contextlib
import pathlib
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

from alembic.command import downgrade, upgrade
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.runtime.migration import MigrationContext, MigrationInfo
from alembic.script import ScriptDirectory
from sqlalchemy import MetaData, String, column, desc, insert, inspect, select, table
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session

from aiida.common import exceptions
from aiida.manage.configuration.profile import Profile
from aiida.storage.log import MIGRATE_LOGGER
from aiida.storage.psql_dos.models.settings import DbSetting
from aiida.storage.psql_dos.utils import create_sqlalchemy_engine

if TYPE_CHECKING:
    from disk_objectstore import Container

TEMPLATE_LEGACY_DJANGO_SCHEMA = """
Database schema is using the legacy Django schema.
To migrate the database schema version to the current one, run the following command:

    verdi -p {profile_name} storage migrate
"""

TEMPLATE_INVALID_SCHEMA_VERSION = """
Database schema version `{schema_version_database}` is incompatible with the required schema version `{schema_version_code}`.
To migrate the database schema version to the current one, run the following command:

    verdi -p {profile_name} storage migrate
"""  # noqa: E501

ALEMBIC_REL_PATH = 'migrations'

REPOSITORY_UUID_KEY = 'repository|uuid'


class PsqlDosMigrator:
    """Class for validating and migrating `psql_dos` storage instances.

    .. important:: This class should only be accessed via the storage backend class (apart from for test purposes)
    """

    alembic_version_tbl_name = 'alembic_version'
    django_version_table = table(
        'django_migrations', column('id'), column('app', String(255)), column('name', String(255)), column('applied')
    )

    def __init__(self, profile: Profile) -> None:
        self.profile = profile
        self._engine = create_sqlalchemy_engine(self.profile.storage_config)
        self._connection = None

    def close(self) -> None:
        """Close the connection if it was opened and dispose of the engine."""
        if self._connection:
            self._connection.close()
            self._connection = None

        if self._engine:
            self._engine.dispose()
            self._engine = None

    @property
    def connection(self):
        """Return the connection to the database.

        Will automatically create the engine and open an connection if not already opened in a previous call.

        :return: Open connection to the database.
        :raises: :class:`aiida.common.exceptions.UnreachableStorage` if connecting to the database fails.
        """
        if self._connection is None:
            try:
                self._connection = self._engine.connect()
            except OperationalError as exception:
                raise exceptions.UnreachableStorage(f'Could not connect to database: {exception}') from exception

        return self._connection

    @classmethod
    def get_schema_versions(cls) -> Dict[str, str]:
        """Return all available schema versions (oldest to latest).

        :return: schema version -> description
        """
        return {entry.revision: entry.doc for entry in reversed(list(cls._alembic_script().walk_revisions()))}

    @classmethod
    def get_schema_version_head(cls) -> str:
        """Return the head schema version for this storage, i.e. the latest schema this storage can be migrated to."""
        return cls._alembic_script().revision_map.get_current_head('main')

    def get_schema_version_profile(self, check_legacy=False) -> Optional[str]:
        """Return the schema version of the backend instance for this profile.

        Note, the version will be None if the database is empty or is a legacy django database.
        """
        with self._migration_context() as context:
            version = context.get_current_revision()
        if version is None and check_legacy:
            stmt = select(self.django_version_table.c.name).where(self.django_version_table.c.app == 'db')
            stmt = stmt.order_by(desc(self.django_version_table.c.id)).limit(1)
            try:
                return self.connection.execute(stmt).scalar()
            except (OperationalError, ProgrammingError):
                self.connection.rollback()
        return version

    def validate_storage(self) -> None:
        """Validate that the storage for this profile

        1. That the database schema is at the head version, i.e. is compatible with the code API.
        2. That the repository ID is equal to the UUID set in the database

        :raises: :class:`aiida.common.exceptions.UnreachableStorage` if the storage cannot be connected to
        :raises: :class:`aiida.common.exceptions.IncompatibleStorageSchema`
            if the storage is not compatible with the code API.
        :raises: :class:`aiida.common.exceptions.CorruptStorage`
            if the repository ID is not equal to the UUID set in thedatabase.
        """
        # check there is an alembic_version table from which to get the schema version
        if not inspect(self.connection).has_table(self.alembic_version_tbl_name):
            # if not present, it might be that this is a legacy django database
            if inspect(self.connection).has_table(self.django_version_table.name):
                raise exceptions.IncompatibleStorageSchema(
                    TEMPLATE_LEGACY_DJANGO_SCHEMA.format(profile_name=self.profile.name)
                )
            raise exceptions.IncompatibleStorageSchema('The database has no known version.')

        # now we can check that the alembic version is the latest
        schema_version_code = self.get_schema_version_head()
        schema_version_database = self.get_schema_version_profile(check_legacy=False)
        if schema_version_database != schema_version_code:
            raise exceptions.IncompatibleStorageSchema(
                TEMPLATE_INVALID_SCHEMA_VERSION.format(
                    schema_version_database=schema_version_database,
                    schema_version_code=schema_version_code,
                    profile_name=self.profile.name,
                )
            )

        # finally, we check that the ID set within the disk-objectstore is equal to the one saved in the database,
        # i.e. this container is indeed the one associated with the db
        repository_uuid = self.get_repository_uuid()
        stmt = select(DbSetting.val).where(DbSetting.key == REPOSITORY_UUID_KEY)
        database_repository_uuid = self.connection.execute(stmt).scalar_one_or_none()
        if database_repository_uuid is None:
            raise exceptions.CorruptStorage('The database has no repository UUID set.')
        if database_repository_uuid != repository_uuid:
            raise exceptions.CorruptStorage(
                f'The database has a repository UUID configured to {database_repository_uuid} '
                f"but the disk-objectstore's is {repository_uuid}."
            )

    def get_container(self) -> 'Container':
        """Return the disk-object store container.

        :returns: The disk-object store container configured for the repository path of the current profile.
        """
        from disk_objectstore import Container

        from .backend import get_filepath_container

        return Container(get_filepath_container(self.profile))

    def get_repository_uuid(self) -> str:
        """Return the UUID of the repository.

        :returns: The repository UUID.
        :raises: :class:`~aiida.common.exceptions.UnreachableStorage` if the UUID cannot be retrieved, which probably
            means that the repository is not initialised.
        """
        try:
            return self.get_container().container_id
        except Exception as exception:
            raise exceptions.UnreachableStorage(
                f'Could not access disk-objectstore {self.get_container()}: {exception}'
            ) from exception

    def initialise(self, reset: bool = False) -> bool:
        """Initialise the storage backend.

        This is typically used once when a new storage backed is created. If this method returns without exceptions the
        storage backend is ready for use. If the backend already seems initialised, this method is a no-op.

        :param reset: If ``true``, destroy the backend if it already exists including all of its data before recreating
            and initialising it. This is useful for example for test profiles that need to be reset before or after
            tests having run.
        :returns: ``True`` if the storage was initialised by the function call, ``False`` if it was already initialised.
        """
        if reset:
            self.reset_repository()
            self.reset_database()

        initialised: bool = False

        if not self.is_initialised:
            self.initialise_repository()
            self.initialise_database()
            initialised = True

        # Call migrate in the case the storage was already initialised but not yet at the latest schema version. If it
        # was, then the following is a no-op anyway.
        self.migrate()

        return initialised

    @property
    def is_initialised(self) -> bool:
        """Return whether the storage is initialised.

        This is the case if both the database and the repository are initialised.

        :returns: ``True`` if the storage is initialised, ``False`` otherwise.
        """
        return self.is_repository_initialised and self.is_database_initialised

    @property
    def is_repository_initialised(self) -> bool:
        """Return whether the repository is initialised.

        :returns: ``True`` if the repository is initialised, ``False`` otherwise.
        """
        return self.get_container().is_initialised

    @property
    def is_database_initialised(self) -> bool:
        """Return whether the database is initialised.

        This is the case if it contains the table that holds the schema version for alembic or Django.

        :returns: ``True`` if the database is initialised, ``False`` otherwise.
        """
        return inspect(self.connection).has_table(self.alembic_version_tbl_name) or inspect(self.connection).has_table(
            self.django_version_table.name
        )

    def reset_repository(self) -> None:
        """Reset the repository by deleting all of its contents.

        This will also destroy the configuration and so in order to use it again, it will have to be reinitialised.
        """
        import shutil

        try:
            shutil.rmtree(self.get_container().get_folder())
        except FileNotFoundError:
            pass

    def reset_database(self) -> None:
        """Reset the database by deleting all content from all tables.

        This will also destroy the settings table and so in order to use it again, it will have to be reinitialised.
        """
        self.delete_all_tables(exclude_tables=[self.alembic_version_tbl_name])

    def initialise_repository(self) -> None:
        """Initialise the repository."""
        from aiida.storage.psql_dos.backend import CONTAINER_DEFAULTS

        container = self.get_container()
        container.init_container(clear=True, **CONTAINER_DEFAULTS)

    def initialise_database(self) -> None:
        """Initialise the database.

        This assumes that the database has no schema whatsoever and so the initial schema is created directly from the
        models at the current head version without migrating through all of them one by one.
        """
        from aiida.storage.psql_dos.models.base import get_orm_metadata

        # setup the database
        # see: https://alembic.sqlalchemy.org/en/latest/cookbook.html#building-an-up-to-date-database-from-scratch
        MIGRATE_LOGGER.report('initialising empty storage schema')
        get_orm_metadata().create_all(self._engine)

        repository_uuid = self.get_repository_uuid()

        # Create a "sync" between the database and repository, by saving its UUID in the settings table
        # this allows us to validate inconsistencies between the two
        self.connection.execute(
            insert(DbSetting).values(key=REPOSITORY_UUID_KEY, val=repository_uuid, description='Repository UUID')
        )

        # finally, generate the version table, "stamping" it with the most recent revision
        with self._migration_context() as context:
            context.stamp(context.script, 'main@head')
            self.connection.commit()

    def delete_all_tables(self, *, exclude_tables: list[str] | None = None) -> None:
        """Delete all tables of the current database schema.

        The tables are determined dynamically through reflection of the current schema version. Any other tables in the
        database that are not part of the schema should remain unaffected.

        :param exclude_tables: Optional list of table names that should not be deleted.
        """
        exclude_tables = exclude_tables or []

        if inspect(self.connection).has_table(self.alembic_version_tbl_name):
            metadata = MetaData()
            metadata.reflect(bind=self.connection)

            # The ``sorted_tables`` property returns the tables sorted by their foreign-key dependencies, with those
            # that are dependent on others first. Iterate over the list in reverse to ensure that the tables with
            # the independent rows are deleted first.
            for schema_table in reversed(metadata.sorted_tables):
                if schema_table.name in exclude_tables:
                    continue
                self.connection.execute(schema_table.delete())
            self.connection.commit()

    def migrate(self) -> None:
        """Migrate the storage for this profile to the head version.

        :raises: :class:`~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed.
        :raises: :class:`~aiida.common.exceptions.StorageMigrationError` if the storage is not initialised.
        """
        # The database can be in one of a few states:
        # 1. Legacy django database -> we transfer the version to alembic, migrate to the head of the django branch,
        #    reset the revision as one on the main branch, and then migrate to the head of the main branch
        # 2. Legacy sqlalchemy database -> we migrate to the head of the sqlalchemy branch,
        #    reset the revision as one on the main branch, and then migrate to the head of the main branch
        # 3. Already on the main branch -> we migrate to the head of the main branch

        if not inspect(self.connection).has_table(self.alembic_version_tbl_name):
            if not inspect(self.connection).has_table(self.django_version_table.name):
                raise exceptions.StorageMigrationError('storage is uninitialised, cannot migrate.')
            # the database is a legacy django one,
            # so we need to copy the version from the 'django_migrations' table to the 'alembic_version' one
            legacy_version = self.get_schema_version_profile(check_legacy=True)
            if legacy_version is None:
                raise exceptions.StorageMigrationError(
                    'No schema version could be read from the database. '
                    "Check that either the 'alembic_version' or 'django_migrations' tables "
                    'are present and accessible, using e.g. `verdi devel run-sql "SELECT * FROM alembic_version"`'
                )
            # the version should be of the format '00XX_description'
            version = f'django_{legacy_version[:4]}'
            with self._migration_context() as context:
                context.stamp(context.script, version)
                self.connection.commit()
            # now we can continue with the migration as normal
        else:
            version = self.get_schema_version_profile()

        # find what branch the current version is on
        branches = self._alembic_script().revision_map.get_revision(version).branch_labels

        if 'django' in branches or 'sqlalchemy' in branches:
            # migrate up to the top of the respective legacy branches
            if 'django' in branches:
                MIGRATE_LOGGER.report('Migrating to the head of the legacy django branch')
                self.migrate_up('django@head')
            elif 'sqlalchemy' in branches:
                MIGRATE_LOGGER.report('Migrating to the head of the legacy sqlalchemy branch')
                self.migrate_up('sqlalchemy@head')
            # now re-stamp with the comparable revision on the main branch
            with self._migration_context() as context:
                context._ensure_version_table(purge=True)
                context.stamp(context.script, 'main_0001')
                self.connection.commit()

        # finally migrate to the main head revision
        MIGRATE_LOGGER.report('Migrating to the head of the main branch')
        self.migrate_up('main@head')
        self.connection.commit()

    def migrate_up(self, version: str) -> None:
        """Migrate the database up to a specific version.

        :param version: string with schema version to migrate to
        """
        with self._alembic_connect() as config:
            upgrade(config, version)

    def migrate_down(self, version: str) -> None:
        """Migrate the database down to a specific version.

        :param version: string with schema version to migrate to
        """
        with self._alembic_connect() as config:
            downgrade(config, version)

    @staticmethod
    def _alembic_config():
        """Return an instance of an Alembic `Config`."""
        dirpath = pathlib.Path(__file__).resolve().parent
        config = Config()
        config.set_main_option('script_location', str(dirpath / ALEMBIC_REL_PATH))
        return config

    @classmethod
    def _alembic_script(cls):
        """Return an instance of an Alembic `ScriptDirectory`."""
        return ScriptDirectory.from_config(cls._alembic_config())

    @contextlib.contextmanager
    def _alembic_connect(self) -> Iterator[Config]:
        """Context manager to return an instance of an Alembic configuration.

        The profiles's database connection is added in the `attributes` property, through which it can then also be
        retrieved, also in the `env.py` file, which is run when the database is migrated.
        """
        config = self._alembic_config()
        config.attributes['connection'] = self.connection
        config.attributes['aiida_profile'] = self.profile

        def _callback(step: MigrationInfo, **kwargs):
            """Callback to be called after a migration step is executed."""
            from_rev = step.down_revision_ids[0] if step.down_revision_ids else '<base>'
            MIGRATE_LOGGER.report(f'- {from_rev} -> {step.up_revision_id}')

        config.attributes['on_version_apply'] = _callback

        yield config

    @contextlib.contextmanager
    def _migration_context(self) -> Iterator[MigrationContext]:
        """Context manager to return an instance of an Alembic migration context.

        This migration context will have been configured with the current database connection, which allows this context
        to be used to inspect the contents of the database, such as the current revision.
        """
        with self._alembic_connect() as config:
            script = ScriptDirectory.from_config(config)
            with EnvironmentContext(config, script) as context:
                context.configure(context.config.attributes['connection'])
                yield context.get_context()

    # the following are used for migration tests

    @contextlib.contextmanager
    def session(self) -> Iterator[Session]:
        """Context manager to return a session for the database."""
        session = Session(self._engine, future=True)
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def get_current_table(self, table_name: str) -> Any:
        """Return a table instantiated at the correct migration.

        Note that this is obtained by inspecting the database and not by looking into the models file.
        So, special methods possibly defined in the models files/classes are not present.
        """
        base = automap_base()
        base.prepare(autoload_with=self.connection.engine)
        return getattr(base.classes, table_name)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Remove legacy workflows

This is similar to migration django_0032

Revision ID: 1b8ed3425af9
Revises: 3d6190594e19
Create Date: 2019-04-03 17:11:44.073582

"""

from alembic import op

from aiida.storage.psql_dos.migrations.utils.legacy_workflows import export_workflow_data

# revision identifiers, used by Alembic.
revision = '1b8ed3425af9'
down_revision = '3d6190594e19'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    # Clean data
    export_workflow_data(op.get_bind(), op.get_context().opts['aiida_profile'])

    op.drop_table('db_dbworkflowstep_sub_workflows')
    op.drop_table('db_dbworkflowstep_calculations')
    op.drop_table('db_dbworkflowstep')
    op.drop_index('ix_db_dbworkflowdata_aiida_obj_id', table_name='db_dbworkflowdata')
    op.drop_index('ix_db_dbworkflowdata_parent_id', table_name='db_dbworkflowdata')
    op.drop_table('db_dbworkflowdata')
    op.drop_index('ix_db_dbworkflow_label', table_name='db_dbworkflow')
    op.drop_table('db_dbworkflow')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Removal of legacy workflows is not reversible.')



"""Fixtures that provides ORM instances."""

from __future__ import annotations

import pathlib
import typing as t

import pytest

if t.TYPE_CHECKING:
    from aiida.orm import Computer


@pytest.fixture(scope='session')
def ssh_key(tmp_path_factory) -> t.Generator[pathlib.Path, None, None]:
    """Generate a temporary SSH key pair for the test session and return the filepath of the private key.

    The filepath of the public key is the same as the private key, but it adds the ``.pub`` file extension.

    :returns: The filepath of the generated private key.
    """
    from uuid import uuid4

    from cryptography.hazmat.backends import default_backend as crypto_default_backend
    from cryptography.hazmat.primitives import serialization as crypto_serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048,
    )

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption(),
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH,
    )

    dirpath = tmp_path_factory.mktemp('keys')
    filename = uuid4().hex
    filepath_private_key = dirpath / filename
    filepath_public_key = dirpath / f'{filename}.pub'

    filepath_private_key.write_bytes(private_key)
    filepath_public_key.write_bytes(public_key)

    try:
        yield filepath_private_key
    finally:
        filepath_private_key.unlink(missing_ok=True)
        filepath_public_key.unlink(missing_ok=True)


@pytest.fixture
def aiida_computer(tmp_path) -> t.Callable[[], 'Computer']:
    """Return a factory to create a new or load an existing :class:`aiida.orm.computers.Computer` instance.

    The database is queried for an existing computer with the same ``label``, ``hostname``, ``scheduler_type`` and
    ``transport_type``. If it exists, it means it was probably created by this fixture in a previous call and it is
    simply returned. Otherwise a new instance is created. Note that the computer is not explicitly configured, unless
    ``configure_kwargs`` are specified. By default the ``localhost`` hostname is used with the ``core.direct`` and
    ``core.local`` scheduler and transport plugins.

    The factory has the following signature:

    :param label: The computer label. If not specified, a random UUID4 is used.
    :param hostname: The hostname of the computer. Defaults to ``localhost``.
    :param scheduler_type: The scheduler plugin to use. Defaults to ``core.direct``.
    :param transport_type: The transport plugin to use. Defaults to ``core.local``.
    :param minimum_job_poll_interval: The default minimum job poll interval to set. Defaults to 0.
    :param default_mpiprocs_per_machine: The default number of MPI procs to set. Defaults to 1.
    :param configuration_kwargs: Optional keyword arguments that, if defined, are used to configure the computer
        by calling :meth:`aiida.orm.computers.Computer.configure`.
    :return: A stored computer instance.
    """

    def factory(
        label: str | None = None,
        hostname='localhost',
        scheduler_type='core.direct',
        transport_type='core.local',
        minimum_job_poll_interval: int = 0,
        default_mpiprocs_per_machine: int = 1,
        configuration_kwargs: dict[t.Any, t.Any] | None = None,
    ) -> 'Computer':
        import uuid

        from aiida.common.exceptions import NotExistent
        from aiida.orm import Computer

        label = label or f'test-computer-{uuid.uuid4().hex}'

        try:
            computer = Computer.collection.get(
                label=label, hostname=hostname, scheduler_type=scheduler_type, transport_type=transport_type
            )
        except NotExistent:
            computer = Computer(
                label=label,
                hostname=hostname,
                workdir=str(tmp_path),
                transport_type=transport_type,
                scheduler_type=scheduler_type,
            )
            computer.store()
            computer.set_minimum_job_poll_interval(minimum_job_poll_interval)
            computer.set_default_mpiprocs_per_machine(default_mpiprocs_per_machine)

        if configuration_kwargs:
            computer.configure(**configuration_kwargs)

        return computer

    return factory


@pytest.fixture
def aiida_computer_local(aiida_computer) -> t.Callable[[], Computer]:
    """Factory to return a :class:`aiida.orm.computers.Computer` instance with ``core.local`` transport.

    Usage::

        def test(aiida_computer_ssh):
            computer = aiida_computer_ssh(label='some-label', configure=True)
            assert computer.transport_type == 'core.local'
            assert computer.is_configured

    The factory has the following signature:

    :param label: The computer label. If not specified, a random UUID4 is used.
    :param configure: Boolean, if ``True``, ensures the computer is configured, otherwise the computer is returned
        as is. Note that if a computer with the given label already exists and it was configured before, the
        computer will not be "un-"configured. If an unconfigured computer is absolutely required, make sure to first
        delete the existing computer or specify another label.
    :return: A stored computer instance.
    """

    def factory(label: str | None = None, configure: bool = True) -> Computer:
        computer = aiida_computer(label=label, hostname='localhost', transport_type='core.local')

        if configure:
            computer.configure()

        return computer

    return factory


@pytest.fixture
def aiida_computer_ssh(aiida_computer, ssh_key) -> t.Callable[[], 'Computer']:
    """Factory to return a :class:`aiida.orm.computers.Computer` instance with ``core.ssh`` transport.

    If ``configure=True``, an SSH key pair is automatically added to the ``.ssh`` folder of the user, allowing an
    actual SSH connection to be made to the localhost.

    Usage::

        def test(aiida_computer_ssh):
            computer = aiida_computer_ssh(label='some-label', configure=True)
            assert computer.transport_type == 'core.ssh'
            assert computer.is_configured

    The factory has the following signature:

    :param label: The computer label. If not specified, a random UUID4 is used.
    :param configure: Boolean, if ``True``, ensures the computer is configured, otherwise the computer is returned
        as is. Note that if a computer with the given label already exists and it was configured before, the
        computer will not be "un-"configured. If an unconfigured computer is absolutely required, make sure to first
        delete the existing computer or specify another label.
    :return: A stored computer instance.
    """

    def factory(label: str | None = None, configure: bool = True) -> 'Computer':
        computer = aiida_computer(label=label, hostname='localhost', transport_type='core.ssh')

        if configure:
            computer.configure(
                key_filename=str(ssh_key),
                key_policy='AutoAddPolicy',
            )

        return computer

    return factory


@pytest.fixture
def aiida_localhost(aiida_computer_local) -> 'Computer':
    """Return a :class:`aiida.orm.computers.Computer` instance representing localhost with ``core.local`` transport.

    Usage::

        def test(aiida_localhost):
            assert aiida_localhost.transport_type == 'core.local'

    :return: The computer.
    """
    return aiida_computer_local(label='localhost')


@pytest.fixture
def aiida_code():
    """Return a factory to create a new or load an existing :class:`aiida.orm.nodes.data.code.abstract.AbstractCode`.

    Usage::

        def test(aiida_localhost, aiida_code):
            from aiida.orm import InstalledCode
            code = aiida_code(
                'core.code.installed',
                label='test-code',
                computer=aiida_localhost,
                filepath_executable='/bin/bash'
            )
            assert isinstance(code, InstalledCode)

    The factory has the following signature:

    :param entry_point: Entry point of the code plugin.
    :param label: The label of the code. Default to a randomly generated string.
    :param kwargs: Additional keyword arguments that are passed to the code's constructor.
    :return: The created or loaded code instance.
    """

    def factory(entry_point: str, label: str | None = None, **kwargs):
        import uuid

        from aiida.common.exceptions import MultipleObjectsError, NotExistent
        from aiida.orm import QueryBuilder
        from aiida.plugins import DataFactory

        cls = DataFactory(entry_point)
        label = label or f'test-code-{uuid.uuid4().hex}'

        try:
            code = QueryBuilder().append(cls, filters={'label': label}).one()[0]
        except (MultipleObjectsError, NotExistent):
            code = cls(label=label, **kwargs).store()

        return code

    return factory


@pytest.fixture
def aiida_code_installed(aiida_code, aiida_localhost):
    """Return a factory to create a new or load an existing :class:`aiida.orm.nodes.data.code.installed.InstalledCode`.

    Usage::

        def test(aiida_code_installed):
            from aiida.orm import InstalledCode
            code = aiida_code_installed()
            assert isinstance(code, InstalledCode)

    The factory has the following signature:

    :param label: The label of the code. Default to a randomly generated string.
    :param default_calc_job_plugin: Optional default calcjob plugin to set.
    :param computer: The computer to set. Defaults to localhost computer of the ``aiida_localhost`` fixture.
    :param filepath_executable: The filepath of the executable. Defaults to ``/bin/bash``.
    :param use_double_quotes: Whether the executable and arguments of the code in the submission script should be
        escaped with single or double quotes.
    :param with_mpi: Whether the executable should be run as an MPI program.
    :param prepend_text: Optional bash commands that should be executed in the submission script before the executable.
    :param append_text: Optional bash commands that should be executed in the submission script after the executable.
    :return: The created or loaded code instance.
    """

    def factory(
        label: str | None = None,
        description: str | None = None,
        default_calc_job_plugin: str | None = None,
        computer: Computer = aiida_localhost,
        filepath_executable: str = '/bin/bash',
        use_double_quotes: bool = False,
        with_mpi: bool | None = None,
        prepend_text: str = '',
        append_text: str = '',
    ):
        return aiida_code(
            'core.code.installed',
            label=label,
            description=description,
            default_calc_job_plugin=default_calc_job_plugin,
            computer=computer,
            filepath_executable=filepath_executable,
            use_double_quotes=use_double_quotes,
            with_mpi=with_mpi,
            prepend_text=prepend_text,
            append_text=append_text,
        )

    return factory



"""Utilities related to ``pydantic``."""

from __future__ import annotations

import typing as t

from pydantic import Field


def MetadataField(  # noqa: N802
    default: t.Any | None = None,
    *,
    priority: int = 0,
    short_name: str | None = None,
    option_cls: t.Any | None = None,
    **kwargs,
):
    """Return a :class:`pydantic.fields.Field` instance with additional metadata.

    .. code-block:: python

        class Model(BaseModel):

            attribute: MetadataField('default', priority=1000, short_name='-A')

    This is a utility function that constructs a ``Field`` instance with an easy interface to add additional metadata.
    It is possible to add metadata using ``Annotated``::

        class Model(BaseModel):

            attribute: Annotated[str, {'metadata': 'value'}] = Field(...)

    However, when requiring multiple metadata, this notation can make the model difficult to read. Since this utility
    is only used to automatically build command line interfaces from the model definition, it is possible to restrict
    which metadata are accepted.

    :param priority: Used to order the list of all fields in the model. Ordering is done from small to large priority.
    :param short_name: Optional short name to use for an option on a command line interface.
    :param option_cls: The :class:`click.Option` class to use to construct the option.
    """
    field_info = Field(default, **kwargs)

    for key, value in (('priority', priority), ('short_name', short_name), ('option_cls', option_cls)):
        if value is not None:
            field_info.metadata.append({key: value})

    return field_info



"""Utilities for logging in the command line interface context."""

import logging

import click

from .echo import COLORS


class CliHandler(logging.Handler):
    """Handler for writing to the console using click."""

    def emit(self, record):
        """Emit log record via click.

        Can make use of special attributes 'nl' (whether to add newline) and 'err' (whether to print to stderr), which
        can be set via the 'extra' dictionary parameter of the logging methods.
        """
        try:
            nl = record.nl
        except AttributeError:
            nl = True

        try:
            err = record.err
        except AttributeError:
            err = False

        try:
            prefix = record.prefix
        except AttributeError:
            prefix = True

        record.prefix = prefix

        try:
            msg = self.format(record)
            click.echo(msg, err=err, nl=nl)
        except Exception:
            self.handleError(record)


class CliFormatter(logging.Formatter):
    """Formatter that automatically prefixes log messages with a colored version of the log level."""

    def format(self, record):
        """Format the record using the style required for the command line interface."""
        try:
            fg = COLORS[record.levelname.lower()]
        except KeyError:
            fg = 'white'

        try:
            prefix = record.prefix
        except AttributeError:
            prefix = None

        formatted = super().format(record)

        if prefix:
            return f'{click.style(record.levelname.capitalize(), fg=fg, bold=True)}: {formatted}'

        return formatted



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data core.cif` command."""

import click

from aiida.cmdline.commands.cmd_data import cmd_show, verdi_data
from aiida.cmdline.commands.cmd_data.cmd_export import data_export, export_options
from aiida.cmdline.commands.cmd_data.cmd_list import data_list, list_options
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo

LIST_PROJECT_HEADERS = ['Id', 'Formulae', 'Source.URI']
EXPORT_FORMATS = ['cif']
VISUALIZATION_FORMATS = ['jmol', 'vesta']


@verdi_data.group('core.cif')
def cif():
    """Manipulate CifData objects (crystal structures in .cif format)."""


@cif.command('list')
@options.FORMULA_MODE()
@list_options
@decorators.with_dbenv()
def cif_list(raw, formula_mode, past_days, groups, all_users):
    """List store CifData objects."""
    from tabulate import tabulate

    from aiida.orm import CifData

    elements = None
    elements_only = False

    entry_list = data_list(
        CifData, LIST_PROJECT_HEADERS, elements, elements_only, formula_mode, past_days, groups, all_users
    )

    counter = 0
    cif_list_data = []

    if not raw:
        cif_list_data.append(LIST_PROJECT_HEADERS)
    for entry in entry_list:
        for i, value in enumerate(entry):
            if isinstance(value, list):
                new_entry = []
                for elm in value:
                    if elm is None:
                        new_entry.append('')
                    else:
                        new_entry.append(elm)
                entry[i] = ','.join(new_entry)
        for i in range(len(entry), len(LIST_PROJECT_HEADERS)):
            entry.append(None)
        counter += 1
    cif_list_data.extend(entry_list)
    if raw:
        echo.echo(tabulate(cif_list_data, tablefmt='plain'))
    else:
        echo.echo(tabulate(cif_list_data, headers='firstrow'))
        echo.echo(f'\nTotal results: {counter}\n')


@cif.command('show')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:core.cif',)))
@options.VISUALIZATION_FORMAT(type=click.Choice(VISUALIZATION_FORMATS), default='jmol')
@decorators.with_dbenv()
def cif_show(data, fmt):
    """Visualize CifData objects."""
    try:
        show_function = getattr(cmd_show, f'_show_{fmt}')
    except AttributeError:
        echo.echo_critical(f'visualization format {fmt} is not supported')

    show_function(fmt, data)


@cif.command('content')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:core.cif',)))
@decorators.with_dbenv()
def cif_content(data):
    """Show the content of the CIF file."""
    for node in data:
        try:
            echo.echo(node.get_content())
        except OSError as exception:
            echo.echo_warning(f'could not read the content for CifData<{node.pk}>: {exception!s}')


@cif.command('export')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:core.cif',)))
@options.EXPORT_FORMAT(type=click.Choice(EXPORT_FORMATS), default='cif')
@export_options
@decorators.with_dbenv()
def cif_export(**kwargs):
    """Export CifData object."""
    node = kwargs.pop('datum')
    output = kwargs.pop('output')
    fmt = kwargs.pop('fmt')
    force = kwargs.pop('force')

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    data_export(node, output, fmt, other_args=kwargs, overwrite=force)


@cif.command('import')
@click.argument('filename', type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@decorators.with_dbenv()
def cif_import(filename):
    """Import .cif file into CifData object."""
    from aiida.orm import CifData

    try:
        node, _ = CifData.get_or_create(filename)
        echo.echo_success(f'imported {node!s}')
    except ValueError as err:
        echo.echo_critical(str(err))



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module to manage comments for the SQLA backend."""

from sqlalchemy import Column, ForeignKey
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import relationship
from sqlalchemy.types import DateTime, Integer, Text

from aiida.common import timezone
from aiida.common.utils import get_new_uuid
from aiida.storage.psql_dos.models.base import Base


class DbComment(Base):
    """Database model to store data for :py:class:`aiida.orm.Comment`.

    Comments can be attach to the nodes by the users.
    """

    __tablename__ = 'db_dbcomment'

    id = Column(Integer, primary_key=True)
    uuid = Column(UUID(as_uuid=True), default=get_new_uuid, nullable=False, unique=True)
    dbnode_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True,
    )
    ctime = Column(DateTime(timezone=True), default=timezone.now, nullable=False)
    mtime = Column(DateTime(timezone=True), default=timezone.now, onupdate=timezone.now, nullable=False)
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True,
    )
    content = Column(Text, default='', nullable=False)

    dbnode = relationship('DbNode', backref='dbcomments')
    user = relationship('DbUser')

    def __str__(self):
        return 'DbComment for [{} {}] on {}'.format(
            self.dbnode.get_simple_name(), self.dbnode.id, timezone.localtime(self.ctime).strftime('%Y-%m-%d')
        )



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration from v0.4 to v0.5, used by `verdi export migrate` command.

The migration steps are named similarly to the database migrations for Django and SQLAlchemy.
In the description of each migration, a revision number is given, which refers to the Django migrations.
The individual Django database migrations may be found at:

    `aiida.storage.djsite.db.migrations.00XX_<migration-name>.py`

Where XX are the numbers in the migrations' documentation: REV. 1.0.XX
And migration-name is the name of the particular migration.
The individual SQLAlchemy database migrations may be found at:

    `aiida.storage.psql_dos.migrations.versions.<id>_<migration-name>.py`

Where id is a SQLA id and migration-name is the name of the particular migration.
"""

from ..utils import update_metadata, verify_metadata_version


def remove_fields(metadata, data, entities, fields):
    """Remove fields under entities from data.json and metadata.json.

    :param metadata: the content of an export archive metadata.json file
    :param data: the content of an export archive data.json file
    :param entities: list of ORM entities
    :param fields: list of fields to be removed from the export archive files
    """
    # data.json
    for entity in entities:
        for content in data['export_data'].get(entity, {}).values():
            for field in fields:
                content.pop(field, None)

    # metadata.json
    for entity in entities:
        for field in fields:
            metadata['all_fields_info'][entity].pop(field, None)


def migration_drop_node_columns_nodeversion_public(metadata, data):
    """Apply migration 0034 - REV. 1.0.34
    Drop the columns `nodeversion` and `public` from the `Node` model
    """
    entity = 'Node'
    fields = ['nodeversion', 'public']

    remove_fields(metadata, data, [entity], fields)


def migration_drop_computer_transport_params(metadata, data):
    """Apply migration 0036 - REV. 1.0.36
    Drop the column `transport_params` from the `Computer` model
    """
    entity = 'Computer'
    field = 'transport_params'

    remove_fields(metadata, data, [entity], [field])


def migrate_v4_to_v5(metadata: dict, data: dict) -> None:
    """Migration of archive files from v0.4 to v0.5

    This is from migration 0034 (drop_node_columns_nodeversion_public) and onwards
    """
    old_version = '0.4'
    new_version = '0.5'

    verify_metadata_version(metadata, old_version)
    update_metadata(metadata, new_version)

    # Apply migrations
    migration_drop_node_columns_nodeversion_public(metadata, data)
    migration_drop_computer_transport_params(metadata, data)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Replace null values with defaults

Revision ID: main_0000a
Revises: main_0000
Create Date: 2022-03-04

"""

import sqlalchemy as sa
from alembic import op

from aiida.common import timezone

# revision identifiers, used by Alembic.
revision = 'main_0000a'
down_revision = 'main_0000'
branch_labels = None
depends_on = None


def upgrade():
    """Convert null values to default values.

    This migration is performed in preparation for the next migration,
    which will make these fields non-nullable.
    """
    db_dbauthinfo = sa.sql.table(
        'db_dbauthinfo',
        sa.sql.column('aiidauser_id', sa.Integer),
        sa.sql.column('dbcomputer_id', sa.Integer),
        sa.Column('enabled', sa.Boolean),
        sa.Column('auth_params', sa.JSON),
        sa.Column('metadata', sa.JSON()),
    )

    # remove rows with null values, which may have previously resulted from deletion of a user or computer
    op.execute(db_dbauthinfo.delete().where(db_dbauthinfo.c.aiidauser_id.is_(None)))
    op.execute(db_dbauthinfo.delete().where(db_dbauthinfo.c.dbcomputer_id.is_(None)))

    op.execute(db_dbauthinfo.update().where(db_dbauthinfo.c.enabled.is_(None)).values(enabled=True))
    op.execute(db_dbauthinfo.update().where(db_dbauthinfo.c.auth_params.is_(None)).values(auth_params={}))
    op.execute(db_dbauthinfo.update().where(db_dbauthinfo.c.metadata.is_(None)).values(metadata={}))

    db_dbcomment = sa.sql.table(
        'db_dbcomment',
        sa.sql.column('dbnode_id', sa.Integer),
        sa.sql.column('user_id', sa.Integer),
        sa.Column('content', sa.Text),
        sa.Column('ctime', sa.DateTime(timezone=True)),
        sa.Column('mtime', sa.DateTime(timezone=True)),
        sa.Column('uuid', sa.CHAR(32)),
    )

    # remove rows with null values, which may have previously resulted from deletion of a node or user
    op.execute(db_dbcomment.delete().where(db_dbcomment.c.dbnode_id.is_(None)))
    op.execute(db_dbcomment.delete().where(db_dbcomment.c.user_id.is_(None)))

    op.execute(db_dbcomment.update().where(db_dbcomment.c.content.is_(None)).values(content=''))
    op.execute(db_dbcomment.update().where(db_dbcomment.c.ctime.is_(None)).values(ctime=timezone.now()))
    op.execute(db_dbcomment.update().where(db_dbcomment.c.mtime.is_(None)).values(mtime=timezone.now()))

    db_dbcomputer = sa.sql.table(
        'db_dbcomputer',
        sa.Column('description', sa.Text),
        sa.Column('hostname', sa.String(255)),
        sa.Column('metadata', sa.JSON()),
        sa.Column('scheduler_type', sa.String(255)),
        sa.Column('transport_type', sa.String(255)),
        sa.Column('uuid', sa.CHAR(32)),
    )

    op.execute(db_dbcomputer.update().where(db_dbcomputer.c.description.is_(None)).values(description=''))
    op.execute(db_dbcomputer.update().where(db_dbcomputer.c.hostname.is_(None)).values(hostname=''))
    op.execute(db_dbcomputer.update().where(db_dbcomputer.c.metadata.is_(None)).values(metadata={}))
    op.execute(db_dbcomputer.update().where(db_dbcomputer.c.scheduler_type.is_(None)).values(scheduler_type=''))
    op.execute(db_dbcomputer.update().where(db_dbcomputer.c.transport_type.is_(None)).values(transport_type=''))

    db_dbgroup = sa.sql.table(
        'db_dbgroup',
        sa.Column('description', sa.Text),
        sa.Column('label', sa.String(255)),
        sa.Column('time', sa.DateTime(timezone=True)),
        sa.Column('type_string', sa.String(255)),
        sa.Column('uuid', sa.CHAR(32)),
    )

    op.execute(db_dbgroup.update().where(db_dbgroup.c.description.is_(None)).values(description=''))
    op.execute(db_dbgroup.update().where(db_dbgroup.c.time.is_(None)).values(time=timezone.now()))
    op.execute(db_dbgroup.update().where(db_dbgroup.c.type_string.is_(None)).values(type_string='core'))

    db_dblog = sa.sql.table(
        'db_dblog',
        sa.Column('levelname', sa.String(255)),
        sa.Column('loggername', sa.String(255)),
        sa.Column('message', sa.Text),
        sa.Column('metadata', sa.JSON()),
        sa.Column('time', sa.DateTime(timezone=True)),
        sa.Column('uuid', sa.CHAR(32)),
    )

    op.execute(db_dblog.update().where(db_dblog.c.levelname.is_(None)).values(levelname=''))
    op.execute(db_dblog.update().where(db_dblog.c.loggername.is_(None)).values(loggername=''))
    op.execute(db_dblog.update().where(db_dblog.c.message.is_(None)).values(message=''))
    op.execute(db_dblog.update().where(db_dblog.c.metadata.is_(None)).values(metadata={}))
    op.execute(db_dblog.update().where(db_dblog.c.time.is_(None)).values(time=timezone.now()))

    db_dbnode = sa.sql.table(
        'db_dbnode',
        sa.Column('ctime', sa.DateTime(timezone=True)),
        sa.Column('description', sa.Text),
        sa.Column('label', sa.String(255)),
        sa.Column('mtime', sa.DateTime(timezone=True)),
        sa.Column('node_type', sa.String(255)),
        sa.Column('uuid', sa.CHAR(32)),
    )

    op.execute(db_dbnode.update().where(db_dbnode.c.ctime.is_(None)).values(ctime=timezone.now()))
    op.execute(db_dbnode.update().where(db_dbnode.c.description.is_(None)).values(description=''))
    op.execute(db_dbnode.update().where(db_dbnode.c.label.is_(None)).values(label=''))
    op.execute(db_dbnode.update().where(db_dbnode.c.mtime.is_(None)).values(mtime=timezone.now()))

    db_dbuser = sa.sql.table(
        'db_dbuser',
        sa.Column('email', sa.String(254)),
        sa.Column('first_name', sa.String(254)),
        sa.Column('last_name', sa.String(254)),
        sa.Column('institution', sa.String(254)),
    )

    op.execute(db_dbuser.update().where(db_dbuser.c.first_name.is_(None)).values(first_name=''))
    op.execute(db_dbuser.update().where(db_dbuser.c.last_name.is_(None)).values(last_name=''))
    op.execute(db_dbuser.update().where(db_dbuser.c.institution.is_(None)).values(institution=''))


def downgrade():
    """Downgrade database schema."""
    raise NotImplementedError('Downgrade of main_0000a.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi status` command."""

import enum
import sys

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import options
from aiida.cmdline.utils import echo
from aiida.common.exceptions import CorruptStorage, IncompatibleStorageSchema, UnreachableStorage
from aiida.common.log import override_log_level
from aiida.common.warnings import warn_deprecation

from ..utils.echo import ExitCode


class ServiceStatus(enum.IntEnum):
    """Describe status of services for 'verdi status' command."""

    UP = 0
    ERROR = 1
    WARNING = 2
    DOWN = 3


STATUS_SYMBOLS = {
    ServiceStatus.UP: {
        'color': 'green',
        'string': '\u2714',
    },
    ServiceStatus.ERROR: {
        'color': 'red',
        'string': '\u2718',
    },
    ServiceStatus.WARNING: {
        'color': 'yellow',
        'string': '\u23fa',
    },
    ServiceStatus.DOWN: {
        'color': 'red',
        'string': '\u2718',
    },
}


@verdi.command('status')
@options.PRINT_TRACEBACK()
@click.option('--no-rmq', is_flag=True, help='Do not check RabbitMQ status')
def verdi_status(print_traceback, no_rmq):
    """Print status of AiiDA services."""
    from aiida import __version__
    from aiida.common.exceptions import ConfigurationError
    from aiida.engine.daemon.client import DaemonException, DaemonNotRunningException
    from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER
    from aiida.manage.manager import get_manager

    exit_code = ExitCode.SUCCESS

    print_status(ServiceStatus.UP, 'version', f'AiiDA v{__version__}')
    print_status(ServiceStatus.UP, 'config', AIIDA_CONFIG_FOLDER)

    manager = get_manager()

    try:
        profile = manager.get_profile()

        if profile is None:
            print_status(ServiceStatus.WARNING, 'profile', 'no profile configured yet')
            echo.echo_report(
                'Run `verdi presto` to automatically setup a profile using all defaults or use `verdi profile setup` '
                'for more control.'
            )
            return

        print_status(ServiceStatus.UP, 'profile', profile.name)

    except Exception as exc:
        message = 'Unable to read AiiDA profile'
        print_status(ServiceStatus.ERROR, 'profile', message, exception=exc, print_traceback=print_traceback)
        sys.exit(ExitCode.CRITICAL)  # stop here - without a profile we cannot access anything

    # Check the backend storage
    storage_head_version = None
    try:
        with override_log_level():  # temporarily suppress noisy logging
            storage_cls = profile.storage_cls
            storage_head_version = storage_cls.version_head()
            storage_backend = storage_cls(profile)
    except UnreachableStorage as exc:
        message = "Unable to connect to profile's storage."
        print_status(ServiceStatus.DOWN, 'storage', message, exception=exc, print_traceback=print_traceback)
        exit_code = ExitCode.CRITICAL
    except IncompatibleStorageSchema:
        message = (
            f'Storage schema version is incompatible with the code version {storage_head_version!r}. '
            'Run `verdi storage migrate` to solve this.'
        )
        print_status(ServiceStatus.DOWN, 'storage', message)
        exit_code = ExitCode.CRITICAL
    except CorruptStorage as exc:
        message = 'Storage is corrupted.'
        print_status(ServiceStatus.DOWN, 'storage', message, exception=exc, print_traceback=print_traceback)
        exit_code = ExitCode.CRITICAL
    except Exception as exc:
        message = "Unable to instatiate profile's storage."
        print_status(ServiceStatus.ERROR, 'storage', message, exception=exc, print_traceback=print_traceback)
        exit_code = ExitCode.CRITICAL
    else:
        message = str(storage_backend)
        print_status(ServiceStatus.UP, 'storage', message)

    if no_rmq:
        warn_deprecation(
            'The `--no-rmq` option is deprecated. If RabbitMQ is not available, a profile should be configured that '
            'sets the `process_control.backend` attribute to `None`.',
            version=3,
        )

    # Getting the broker status
    broker = manager.get_broker()

    if broker:
        try:
            broker.get_communicator()
        except Exception as exc:
            message = f'Unable to connect to broker: {broker}'
            print_status(ServiceStatus.ERROR, 'broker', message, exception=exc, print_traceback=print_traceback)
            exit_code = ExitCode.CRITICAL
        else:
            print_status(ServiceStatus.UP, 'broker', broker)
    else:
        print_status(
            ServiceStatus.WARNING,
            'broker',
            'No broker defined for this profile: certain functionality not available.',
        )

    # Getting the daemon status
    try:
        status = manager.get_daemon_client().get_status()
    except ConfigurationError:
        print_status(
            ServiceStatus.WARNING,
            'daemon',
            'No broker defined for this profile: daemon is not available.',
        )
    except DaemonNotRunningException as exception:
        print_status(ServiceStatus.WARNING, 'daemon', str(exception))
    except DaemonException as exception:
        print_status(ServiceStatus.ERROR, 'daemon', str(exception))
    except Exception as exception:
        message = 'Error getting daemon status'
        print_status(ServiceStatus.ERROR, 'daemon', message, exception=exception, print_traceback=print_traceback)
        exit_code = ExitCode.CRITICAL
    else:
        print_status(ServiceStatus.UP, 'daemon', f'Daemon is running with PID {status["pid"]}')

    # Note: click does not forward return values to the exit code, see https://github.com/pallets/click/issues/747
    if exit_code != ExitCode.SUCCESS:
        sys.exit(exit_code)


def print_status(status, service, msg='', exception=None, print_traceback=False):
    """Print status message.

    Includes colored indicator.

    :param status: a ServiceStatus code
    :param service: string for service name
    :param msg:  message string
    """
    symbol = STATUS_SYMBOLS[status]
    echo.echo(f" {symbol['string']} ", fg=symbol['color'], nl=False)
    echo.echo(f"{service + ':':12s} {msg}")

    if exception is not None:
        echo.echo_error(f'{type(exception).__name__}: {exception}')

    if print_traceback:
        import traceback

        traceback.print_exc()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tools and an option class for interactive parameter entry with additional features such as help lookup."""

import typing as t

import click

from aiida.cmdline.utils import echo

from .conditional import ConditionalOption


class InteractiveOption(ConditionalOption):
    """Prompts for input, intercepting certain keyword arguments to provide more feature-rich behavior.

    .. note:: This class has a parameter ``required_fn`` that can be passed to its ``__init__`` (inherited
        from the superclass :py:class:`~aiida.cmdline.params.options.conditional.ConditionalOption`) and a
        ``prompt_fn``.

        - ``required_fn`` is about "is this parameter required" depending on the value of other params.
        - ``prompt_fn`` is about "should I prompt for this value if in interactive mode" and only makes sense
          in this class and not in :py:class:`~aiida.cmdline.params.options.conditional.ConditionalOption`.

        In most usecases, if I have a ``prompt_fn``, then I would like to have also (the same) ``required_fn``.
        The implementation still makes them independent for usecases where they might be different functions
        (e.g. if the variable is anyway not required, but you want to decide whether to prompt for it or not).

    Usage::

        import click

        @click.command()
        @click.option('label', prompt='Label', cls=InteractiveOption)
        def foo(label):
            click.echo(f'Labeling with label: {label}')

    """

    PROMPT_COLOR = echo.COLORS['warning']
    CHARACTER_PROMPT_HELP = '?'
    CHARACTER_IGNORE_DEFAULT = '!'

    def __init__(self, param_decls=None, prompt_fn=None, contextual_default=None, **kwargs):
        """Construct a new instance.

        :param param_decls: relayed to :class:`click.Option`
        :param prompt_fn: callable(ctx) -> Bool, returns True if the option should be prompted for in interactive mode.
        :param contextual_default: An optional callback function to get a default which is passed the click context.
        """
        super().__init__(param_decls=param_decls, **kwargs)
        self._prompt_fn = prompt_fn
        self._contextual_default = contextual_default

    @property
    def prompt(self):
        """Return a colorized version of the prompt text."""
        return click.style(self._prompt, fg=self.PROMPT_COLOR)

    @prompt.setter
    def prompt(self, value):
        """Set the prompt text."""
        self._prompt = value

    def prompt_for_value(self, ctx: click.Context) -> t.Any:
        """Prompt for a value printing a generic help message if this is the first invocation of the command.

        If the command is invoked in non-interactive mode, meaning one should never prompt for a value, the default is
        returned instead of prompting.

        If the help message is printed, the ``prompt_loop_info_printed`` variable is set in the context which is used
        to check whether the message has already been printed as to only print it once at the first prompt.
        """
        if not self.is_interactive(ctx):
            return self.get_default(ctx)

        if self._prompt_fn is not None and self._prompt_fn(ctx) is False:
            return None

        if not hasattr(ctx, 'prompt_loop_info_printed'):
            echo.echo_report(f'enter {self.CHARACTER_PROMPT_HELP} for help.')
            echo.echo_report(f'enter {self.CHARACTER_IGNORE_DEFAULT} to ignore the default and set no value.')
            ctx.prompt_loop_info_printed = True

        return super().prompt_for_value(ctx)

    def process_value(self, ctx: click.Context, value: t.Any) -> t.Any:
        """Intercept any special characters before calling parent class if in interactive mode.

            * If the value matches ``CHARACTER_PROMPT_HELP``, echo ``get_help_message`` and reprompt.
            * If the value matches ``CHARACTER_IGNORE_DEFAULT``, ignore the value and return ``None``.

        Note that this logic only applies if the value is specified at the prompt, if it is provided from the command
        line, the value is actually taken as the value and processed as normal. To determine how the parameter was
        specified the ``click.Context.get_parameter_source`` method is used. The ``click.Parameter.handle_parse_result``
        method will set this after ``Parameter.consume_value``` is called but before ``Parameter.process_value`` is.
        """
        source = ctx.get_parameter_source(self.name)

        if source is None:
            return value

        if value == self.CHARACTER_PROMPT_HELP and source is click.core.ParameterSource.PROMPT:
            click.echo(self.get_help_message())
            return self.prompt_for_value(ctx)

        if value == self.CHARACTER_IGNORE_DEFAULT and source is click.core.ParameterSource.PROMPT:
            # This means the user does not want to set a specific value for this option, so the ``!`` character is
            # translated to ``None`` and processed as normal. If the option is required, a validation error will be
            # raised further down below, forcing the user to specify a valid value.
            value = None

        try:
            return super().process_value(ctx, value)
        except click.BadParameter as exception:
            if source is click.core.ParameterSource.PROMPT and self.is_interactive(ctx):
                if isinstance(exception, click.MissingParameter):
                    click.echo(f'Error: {self._prompt} has to be specified')
                else:
                    click.echo(f'Error: {exception}')
                return self.prompt_for_value(ctx)
            raise

    def get_help_message(self):
        """Return a message to be displayed for in-prompt help."""
        message = self.help or f'Expecting {self.type.name}'

        choices = getattr(self.type, 'shell_complete', lambda x, y, z: [])(self.type, None, '')
        choices_string = []

        for choice in choices:
            if choice.value and choice.help:
                choices_string.append(f' * {choice.value:<12} {choice.help}')
            elif choice.value:
                choices_string.append(f' * {choice.value}')

        if any(choices_string):
            message += '\nSelect one of:\n'
            message += '\n'.join([choice for choice in choices_string if choice.strip()])

        return message

    def get_default(self, ctx: click.Context, call: bool = True) -> t.Optional[t.Union[t.Any, t.Callable[[], t.Any]]]:
        """Provides the functionality of :meth:`click.Option.get_default`"""
        if ctx.resilient_parsing:
            return None

        if self._contextual_default is not None:
            default = self._contextual_default(ctx)
        else:
            default = super().get_default(ctx, call=call)

        try:
            default = self.type.deconvert_default(default)
        except AttributeError:
            pass

        return default

    @staticmethod
    def is_interactive(ctx: click.Context) -> bool:
        """Return whether the command is being run non-interactively.

        This is the case if the ``non_interactive`` parameter in the context is set to ``False``.

        :return: ``True`` if being run interactively, ``False`` otherwise.
        """
        return not ctx.params.get('non_interactive', False)


class TemplateInteractiveOption(InteractiveOption):
    """Sub class of ``InteractiveOption`` that uses template file for input instead of simple inline prompt.

    This is useful for options that need to be able to specify multiline string values.
    """

    def __init__(self, param_decls=None, **kwargs):
        """Define the configuration for the multiline template in the keyword arguments.

        :param template: name of the template to use from the ``aiida.cmdline.templates`` directory.
            Default is the 'multiline.tpl' template.
        :param header: string to put in the header of the template.
        :param footer: string to put in the footer of the template.
        :param extension: file extension to give to the template file.
        """
        self.template = kwargs.pop('template', 'multiline.tpl')
        self.header = kwargs.pop('header', '')
        self.footer = kwargs.pop('footer', '')
        self.extension = kwargs.pop('extension', '')
        super().__init__(param_decls=param_decls, **kwargs)

    def prompt_for_value(self, ctx: click.Context) -> t.Any:
        """Replace the basic prompt with a method that opens a template file in an editor."""
        from aiida.cmdline.utils.multi_line_input import edit_multiline_template

        if not self.is_interactive(ctx):
            return self.get_default(ctx)

        if self._prompt_fn is not None and self._prompt_fn(ctx) is False:
            return None

        kwargs = {'value': self.get_default(ctx) or '', 'header': self.header, 'footer': self.footer}
        return edit_multiline_template(self.template, extension=self.extension, **kwargs)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Generic backend related objects"""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, ContextManager, List, Optional, Sequence, TypeVar, Union

if TYPE_CHECKING:
    from aiida.manage.configuration.profile import Profile
    from aiida.orm.autogroup import AutogroupManager
    from aiida.orm.entities import EntityTypes
    from aiida.orm.implementation import (
        BackendAuthInfoCollection,
        BackendCommentCollection,
        BackendComputerCollection,
        BackendGroupCollection,
        BackendLogCollection,
        BackendNodeCollection,
        BackendQueryBuilder,
        BackendUserCollection,
    )
    from aiida.orm.users import User
    from aiida.repository.backend.abstract import AbstractRepositoryBackend

__all__ = ('StorageBackend',)

TransactionType = TypeVar('TransactionType')


class StorageBackend(abc.ABC):
    """Abstraction for a backend to read/write persistent data for a profile's provenance graph.

    AiiDA splits data storage into two sources:

    - Searchable data, which is stored in the database and can be queried using the QueryBuilder
    - Non-searchable (binary) data, which is stored in the repository and can be loaded using the RepositoryBackend

    The two sources are inter-linked by the ``Node.base.repository.metadata``.
    Once stored, the leaf values of this dictionary must be valid pointers to object keys in the repository.

    For a completely new storage, the ``initialise`` method should be called first. This will automatically initialise
    the repository and the database with the current schema. The class methods,`version_profile` and `migrate` should be
    able to be called for existing storage, at any supported schema version. But an instance of this class should be
    created only for the latest schema version.
    """

    read_only = False

    @classmethod
    @abc.abstractmethod
    def version_head(cls) -> str:
        """Return the head schema version of this storage backend type."""

    @classmethod
    @abc.abstractmethod
    def version_profile(cls, profile: 'Profile') -> Optional[str]:
        """Return the schema version of the given profile's storage, or None for empty/uninitialised storage.

        :raises: `~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed
        """

    @classmethod
    @abc.abstractmethod
    def initialise(cls, profile: 'Profile', reset: bool = False) -> bool:
        """Initialise the storage backend.

        This is typically used once when a new storage backed is created. If this method returns without exceptions the
        storage backend is ready for use. If the backend already seems initialised, this method is a no-op.

        :param reset: If ``true``, destroy the backend if it already exists including all of its data before recreating
            and initialising it. This is useful for example for test profiles that need to be reset before or after
            tests having run.
        :returns: ``True`` if the storage was initialised by the function call, ``False`` if it was already initialised.
        """

    @classmethod
    @abc.abstractmethod
    def migrate(cls, profile: 'Profile') -> None:
        """Migrate the storage of a profile to the latest schema version.

        If the schema version is already the latest version, this method does nothing. If the storage is uninitialised,
        this method will raise an exception.

        :raises: :class`~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed.
        :raises: :class:`~aiida.common.exceptions.StorageMigrationError` if the storage is not initialised.
        """

    @abc.abstractmethod
    def __init__(self, profile: 'Profile') -> None:
        """Initialize the backend, for this profile.

        :raises: `~aiida.common.exceptions.UnreachableStorage` if the storage cannot be accessed
        :raises: `~aiida.common.exceptions.IncompatibleStorageSchema`
            if the profile's storage schema is not at the latest version (and thus should be migrated)
        :raises: :raises: :class:`aiida.common.exceptions.CorruptStorage` if the storage is internally inconsistent
        """
        from aiida.orm.autogroup import AutogroupManager

        self._profile = profile
        self._default_user: Optional['User'] = None
        self._autogroup = AutogroupManager(self)

    @abc.abstractmethod
    def __str__(self) -> str:
        """Return a string showing connection details for this instance."""

    @property
    def profile(self) -> 'Profile':
        """Return the profile for this backend."""
        return self._profile

    @property
    def autogroup(self) -> 'AutogroupManager':
        """Return the autogroup manager for this backend."""
        return self._autogroup

    def version(self) -> str:
        """Return the schema version of the profile's storage."""
        version = self.version_profile(self.profile)
        assert version is not None
        return version

    @abc.abstractmethod
    def close(self):
        """Close the storage access."""

    @property
    @abc.abstractmethod
    def is_closed(self) -> bool:
        """Return whether the storage is closed."""

    @abc.abstractmethod
    def _clear(self) -> None:
        """Clear the storage, removing all data.

        .. warning:: This is a destructive operation, and should only be used for testing purposes.
        """
        from aiida.orm.autogroup import AutogroupManager

        self.reset_default_user()
        self._autogroup = AutogroupManager(self)

    def reset_default_user(self) -> None:
        """Reset the default user.

        This should be done when the default user of the storage backend is changed on the corresponding profile because
        the old default user is cached on this instance.
        """
        self._default_user = None

    @property
    @abc.abstractmethod
    def authinfos(self) -> 'BackendAuthInfoCollection':
        """Return the collection of authorisation information objects"""

    @property
    @abc.abstractmethod
    def comments(self) -> 'BackendCommentCollection':
        """Return the collection of comments"""

    @property
    @abc.abstractmethod
    def computers(self) -> 'BackendComputerCollection':
        """Return the collection of computers"""

    @property
    @abc.abstractmethod
    def groups(self) -> 'BackendGroupCollection':
        """Return the collection of groups"""

    @property
    @abc.abstractmethod
    def logs(self) -> 'BackendLogCollection':
        """Return the collection of logs"""

    @property
    @abc.abstractmethod
    def nodes(self) -> 'BackendNodeCollection':
        """Return the collection of nodes"""

    @property
    @abc.abstractmethod
    def users(self) -> 'BackendUserCollection':
        """Return the collection of users"""

    @property
    def default_user(self) -> Optional['User']:
        """Return the default user for the profile, if it has been created.

        This is cached, since it is a frequently used operation, for creating other entities.
        """
        from aiida.orm import QueryBuilder, User

        if self._default_user is None and self.profile.default_user_email:
            query = QueryBuilder(self).append(User, filters={'email': self.profile.default_user_email})
            self._default_user = query.first(flat=True)
        return self._default_user

    @abc.abstractmethod
    def query(self) -> 'BackendQueryBuilder':
        """Return an instance of a query builder implementation for this backend"""

    @abc.abstractmethod
    def transaction(self) -> ContextManager[Any]:
        """Get a context manager that can be used as a transaction context for a series of backend operations.
        If there is an exception within the context then the changes will be rolled back and the state will
        be as before entering.  Transactions can be nested.

        :return: a context manager to group database operations
        """

    @property
    @abc.abstractmethod
    def in_transaction(self) -> bool:
        """Return whether a transaction is currently active."""

    @abc.abstractmethod
    def bulk_insert(self, entity_type: 'EntityTypes', rows: List[dict], allow_defaults: bool = False) -> List[int]:
        """Insert a list of entities into the database, directly into a backend transaction.

        :param entity_type: The type of the entity
        :param data: A list of dictionaries, containing all fields of the backend model,
            except the `id` field (a.k.a primary key), which will be generated dynamically
        :param allow_defaults: If ``False``, assert that each row contains all fields (except primary key(s)),
            otherwise, allow default values for missing fields.

        :raises: ``IntegrityError`` if the keys in a row are not a subset of the columns in the table

        :returns: The list of generated primary keys for the entities
        """

    @abc.abstractmethod
    def bulk_update(self, entity_type: 'EntityTypes', rows: List[dict]) -> None:
        """Update a list of entities in the database, directly with a backend transaction.

        :param entity_type: The type of the entity
        :param data: A list of dictionaries, containing fields of the backend model to update,
            and the `id` field (a.k.a primary key)

        :raises: ``IntegrityError`` if the keys in a row are not a subset of the columns in the table
        """

    def delete(self) -> None:
        """Delete the storage and all the data."""
        raise NotImplementedError()

    @abc.abstractmethod
    def delete_nodes_and_connections(self, pks_to_delete: Sequence[int]):
        """Delete all nodes corresponding to pks in the input and any links to/from them.

        This method is intended to be used within a transaction context.

        :param pks_to_delete: a sequence of node pks to delete

        :raises: ``AssertionError`` if a transaction is not active
        """

    @abc.abstractmethod
    def get_repository(self) -> 'AbstractRepositoryBackend':
        """Return the object repository configured for this backend."""

    @abc.abstractmethod
    def set_global_variable(
        self, key: str, value: Union[None, str, int, float], description: Optional[str] = None, overwrite=True
    ) -> None:
        """Set a global variable in the storage.

        :param key: the key of the setting
        :param value: the value of the setting
        :param description: the description of the setting (optional)
        :param overwrite: if True, overwrite the setting if it already exists

        :raises: `ValueError` if the key already exists and `overwrite` is False
        """

    @abc.abstractmethod
    def get_global_variable(self, key: str) -> Union[None, str, int, float]:
        """Return a global variable from the storage.

        :param key: the key of the setting

        :raises: `KeyError` if the setting does not exist
        """

    @abc.abstractmethod
    def maintain(self, full: bool = False, dry_run: bool = False, **kwargs) -> None:
        """Perform maintenance tasks on the storage.

        If `full == True`, then this method may attempt to block the profile associated with the
        storage to guarantee the safety of its procedures. This will not only prevent any other
        subsequent process from accessing that profile, but will also first check if there is
        already any process using it and raise if that is the case. The user will have to manually
        stop any processes that is currently accessing the profile themselves or wait for it to
        finish on its own.

        :param full: flag to perform operations that require to stop using the profile to be maintained.
        :param dry_run: flag to only print the actions that would be taken without actually executing them.
        """

    def _backup(
        self,
        dest: str,
        keep: Optional[int] = None,
    ):
        raise NotImplementedError

    def _write_backup_config(self, backup_manager):
        import pathlib
        import tempfile

        from aiida.common import exceptions
        from aiida.common.log import override_log_level
        from aiida.manage.configuration import get_config
        from aiida.manage.configuration.config import Config
        from aiida.manage.configuration.settings import DEFAULT_CONFIG_FILE_NAME

        try:
            config = get_config()
            profile = config.get_profile(self.profile.name)  # Get the profile being backed up
            with tempfile.TemporaryDirectory() as tmpdir:
                filepath_config = pathlib.Path(tmpdir) / DEFAULT_CONFIG_FILE_NAME
                backup_config = Config(str(filepath_config), {})  # Create empty config at temporary file location
                backup_config.add_profile(profile)  # Add the profile being backed up
                backup_config.store()  # Write the contents to disk

                # Temporarily disable all logging because the verbose rsync output just for copying the config file
                # is a bit much.
                with override_log_level():
                    backup_manager.call_rsync(filepath_config, backup_manager.path / DEFAULT_CONFIG_FILE_NAME)
        except (exceptions.MissingConfigurationError, exceptions.ConfigurationError) as exc:
            raise exceptions.StorageBackupError('AiiDA config.json not found!') from exc

    def _validate_or_init_backup_folder(self, dest, keep):
        import json
        import tempfile

        from disk_objectstore import backup_utils

        from aiida.common import exceptions
        from aiida.manage.configuration.config import Config
        from aiida.manage.configuration.settings import DEFAULT_CONFIG_FILE_NAME
        from aiida.storage.log import STORAGE_LOGGER

        try:
            # this creates the dest folder if it doesn't exist
            backup_manager = backup_utils.BackupManager(dest, keep=keep)
            backup_config_path = backup_manager.path / DEFAULT_CONFIG_FILE_NAME

            if backup_manager.check_path_exists(backup_config_path):
                success, stdout = backup_manager.run_cmd(['cat', str(backup_config_path)])
                if not success:
                    raise exceptions.StorageBackupError(f"Couldn't read {backup_config_path!s}.")
                try:
                    backup_config_existing = json.loads(stdout)
                except json.decoder.JSONDecodeError as exc:
                    raise exceptions.StorageBackupError(f'JSON parsing failed for {backup_config_path!s}: {exc.msg}')

                # create a temporary config file to access the profile info
                with tempfile.NamedTemporaryFile() as temp_file:
                    backup_config = Config(temp_file.name, backup_config_existing, validate=False)
                    if len(backup_config.profiles) != 1:
                        raise exceptions.StorageBackupError(f"{backup_config_path!s} doesn't contain exactly 1 profile")

                    if (
                        backup_config.profiles[0].uuid != self.profile.uuid
                        or backup_config.profiles[0].storage_backend != self.profile.storage_backend
                    ):
                        raise exceptions.StorageBackupError(
                            'The chosen destination contains backups of a different profile! Aborting!'
                        )
            else:
                STORAGE_LOGGER.warning('Initializing a new backup folder.')
                # make sure the folder is empty
                success, stdout = backup_manager.run_cmd(['ls', '-A', str(backup_manager.path)])
                if not success:
                    raise exceptions.StorageBackupError(f"Couldn't read {backup_manager.path!s}.")
                if stdout:
                    raise exceptions.StorageBackupError("Can't initialize the backup folder, destination is not empty.")

                self._write_backup_config(backup_manager)

        except backup_utils.BackupError as exc:
            raise exceptions.StorageBackupError(*exc.args) from exc

        return backup_manager

    def backup(
        self,
        dest: str,
        keep: Optional[int] = None,
    ):
        """Create a backup of the storage contents.

        :param dest: The path to the destination folder.
        :param keep: The number of backups to keep in the target destination, if the backend supports it.
        :raises ValueError: If the input parameters are invalid.
        :raises StorageBackupError: If an error occurred during the backup procedure.
        :raises NotImplementedError: If the storage backend doesn't implement a backup procedure.
        """
        from aiida.common.exceptions import LockedProfileError, StorageBackupError
        from aiida.manage.configuration.settings import DEFAULT_CONFIG_FILE_NAME
        from aiida.manage.profile_access import ProfileAccessManager
        from aiida.storage.log import STORAGE_LOGGER

        # check that the AiiDA profile is not locked and request access for the duration of this backup process
        # (locked means that possibly a maintenance operation is running that could interfere with the backup)
        try:
            ProfileAccessManager(self._profile).request_access()
        except LockedProfileError as exc:
            raise StorageBackupError(f'{self._profile} is locked!') from exc

        backup_manager = self._validate_or_init_backup_folder(dest, keep)

        try:
            self._backup(dest, keep)
        except NotImplementedError:
            success, stdout = backup_manager.run_cmd(['ls', '-A', str(backup_manager.path)])

            if not success:
                STORAGE_LOGGER.warning(f'Failed to determine contents of destination folder `{dest}`: not deleting it.')
                raise

            # If the backup directory was just initialized for the first time, it should only contain the configuration
            # file and nothing else. If anything else is found, do not delete the directory for safety reasons.
            if stdout.strip() != DEFAULT_CONFIG_FILE_NAME:
                STORAGE_LOGGER.warning(f'The destination folder `{dest}` is not empty: not deleting it.')
                raise

            backup_manager.run_cmd(['rm', '-rf', str(backup_manager.path)])
            raise

        STORAGE_LOGGER.report(f'Overwriting the `{DEFAULT_CONFIG_FILE_NAME} file.')
        self._write_backup_config(backup_manager)

    def get_info(self, detailed: bool = False) -> dict:
        """Return general information on the storage.

        :param detailed: flag to request more detailed information about the content of the storage.
        :returns: a nested dict with the relevant information.
        """
        return {'entities': self.get_orm_entities(detailed=detailed)}

    def get_orm_entities(self, detailed: bool = False) -> dict:
        """Return a mapping with an overview of the storage contents regarding ORM entities.

        :param detailed: flag to request more detailed information about the content of the storage.
        :returns: a nested dict with the relevant information.
        """
        from aiida.orm import Comment, Computer, Group, Log, Node, QueryBuilder, User

        data = {}

        query_user = QueryBuilder(self).append(User, project=['email'])
        data['Users'] = {'count': query_user.count()}
        if detailed:
            data['Users']['emails'] = sorted({email for (email,) in query_user.iterall() if email is not None})

        query_comp = QueryBuilder(self).append(Computer, project=['label'])
        data['Computers'] = {'count': query_comp.count()}
        if detailed:
            data['Computers']['labels'] = sorted({comp for (comp,) in query_comp.iterall() if comp is not None})

        count = QueryBuilder(self).append(Node).count()
        data['Nodes'] = {'count': count}
        if detailed:
            node_types = sorted(
                {typ for (typ,) in QueryBuilder(self).append(Node, project=['node_type']).iterall() if typ is not None}
            )
            data['Nodes']['node_types'] = node_types
            process_types = sorted(
                {
                    typ
                    for (typ,) in QueryBuilder(self).append(Node, project=['process_type']).iterall()
                    if typ is not None
                }
            )
            data['Nodes']['process_types'] = [p for p in process_types if p]

        query_group = QueryBuilder(self).append(Group, project=['type_string'])
        data['Groups'] = {'count': query_group.count()}
        if detailed:
            data['Groups']['type_strings'] = sorted({typ for (typ,) in query_group.iterall() if typ is not None})

        count = QueryBuilder(self).append(Comment).count()
        data['Comments'] = {'count': count}

        count = QueryBuilder(self).append(Log).count()
        data['Logs'] = {'count': count}

        count = QueryBuilder(self).append(entity_type='link').count()
        data['Links'] = {'count': count}

        return data



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Drop the DbCalcState table

Revision ID: 162b99bca4a2
Revises: a603da2cc809
Create Date: 2018-11-14 08:37:13.719646

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '162b99bca4a2'
down_revision = 'a603da2cc809'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.drop_table('db_dbcalcstate')


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of 162b99bca4a2.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Resources for REST API"""

from urllib.parse import unquote

from flask import make_response, request
from flask_restful import Resource

from aiida.common.lang import classproperty
from aiida.manage import get_config_option, load_profile
from aiida.restapi.common.exceptions import RestInputValidationError
from aiida.restapi.common.utils import Utils, close_thread_connection
from aiida.restapi.translator.nodes.node import NodeTranslator


class ServerInfo(Resource):
    """Endpoint to return general server info"""

    def __init__(self, **kwargs):
        # Configure utils
        utils_conf_keys = ('PREFIX', 'PERPAGE_DEFAULT', 'LIMIT_DEFAULT')
        self.utils_confs = {k: kwargs[k] for k in utils_conf_keys if k in kwargs}
        self.utils = Utils(**self.utils_confs)

    def get(self):
        """It returns the general info about the REST API
        :return: returns current AiiDA version defined in aiida/__init__.py
        """
        path = unquote(request.path)
        url = unquote(request.url)
        url_root = unquote(request.url_root)
        query_string = unquote(request.query_string.decode('utf-8'))
        subpath = self.utils.strip_api_prefix(path).strip('/')
        pathlist = self.utils.split_path(subpath)

        if subpath == '':
            resource_type = 'endpoints'
        elif len(pathlist) > 1:
            resource_type = pathlist.pop(1)
        else:
            resource_type = 'info'

        response = {}

        from aiida import __version__
        from aiida.restapi.common.config import API_CONFIG

        if resource_type == 'info':
            response = {}

            # Add Rest API version
            api_version = API_CONFIG['VERSION'].split('.')
            response['API_major_version'] = api_version[0]
            response['API_minor_version'] = api_version[1]
            response['API_revision_version'] = api_version[2]

            # Add Rest API prefix
            response['API_prefix'] = API_CONFIG['PREFIX']

            # Add AiiDA version
            response['AiiDA_version'] = __version__

        elif resource_type == 'endpoints':
            from aiida.restapi.common.utils import list_routes

            response['available_endpoints'] = list_routes()

        headers = self.utils.build_headers(url=request.url, total_count=1)

        ## Build response and return it
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=path,
            query_string=query_string,
            resource_type='Info',
            data=response,
        )
        return self.utils.build_response(status=200, headers=headers, data=data)


class BaseResource(Resource):
    """Each derived class will instantiate a different type of translator.
    This is the only difference in the classes.
    """

    from aiida.restapi.translator.base import BaseTranslator

    _translator_class = BaseTranslator
    _parse_pk_uuid = None  # Flag to tell the path parser whether to expect a pk or a uuid pattern

    method_decorators = [close_thread_connection]  # Close the thread's storage connection after any method call

    ## TODO add the caching support. I cache total count, results, and possibly

    def __init__(self, profile, **kwargs):
        """Construct the resource."""
        self.trans = self._translator_class(**kwargs)
        self.profile = profile

        # Configure utils
        utils_conf_keys = ('PREFIX', 'PERPAGE_DEFAULT', 'LIMIT_DEFAULT')
        self.utils_confs = {k: kwargs[k] for k in utils_conf_keys if k in kwargs}
        self.utils = Utils(**self.utils_confs)

        # HTTP Request method decorators
        if 'get_decorators' in kwargs and isinstance(kwargs['get_decorators'], (tuple, list, set)):
            self.method_decorators = {'get': list(kwargs['get_decorators'])}

    @staticmethod
    def unquote_request():
        """Unquote and return various parts of the request.

        :returns: Tuple of the request path, url, url root and query string.
        """
        path = unquote(request.path)
        url = unquote(request.url)
        url_root = unquote(request.url_root)
        query_string = unquote(request.query_string.decode('utf-8'))
        return path, url, url_root, query_string

    def parse_path(self, path):
        """Parse the request path."""
        return self.utils.parse_path(path, parse_pk_uuid=self.parse_pk_uuid)

    def parse_query_string(self, query_string):
        """Parse the request query string."""
        return self.utils.parse_query_string(query_string)

    @classproperty
    def parse_pk_uuid(cls):  # noqa: N805
        return cls._parse_pk_uuid

    def _load_and_verify(self, node_id=None):
        """Load node and verify it is of the required type"""
        from aiida.orm import load_node

        node = load_node(node_id)

        if not isinstance(node, self.trans._aiida_class):
            raise RestInputValidationError(f'node {node_id} is not of the required type {self.trans._aiida_class}')

        return node

    def load_profile(self, profile=None):
        """Load the required profile.

        This will load the profile specified by the ``profile`` keyword in the query parameters, and if not specified it
        will default to the profile defined in the constructor.
        """
        profile_switching_enabled = get_config_option('rest_api.profile_switching')

        if profile is not None and not profile_switching_enabled:
            raise RestInputValidationError('specifying an explicit profile is not enabled for this REST API.')

        load_profile(profile or self.profile, allow_switch=True)

    def get(self, id=None, page=None):
        """Get method for the resource
        :param id: node identifier
        :param page: page no, used for pagination
        :return: http response
        """
        path, url, url_root, query_string = self.unquote_request()
        resource_type, page, node_id, query_type = self.parse_path(path)
        parameters = self.parse_query_string(query_string)
        limit = parameters[0]
        offset = parameters[1]
        perpage = parameters[2]
        orderby = parameters[3]
        filters = parameters[4]
        profile = parameters[-1]

        try:
            self.load_profile(profile)
        except RestInputValidationError as exception:
            return self.utils.build_response(
                status=400,
                headers=self.utils.build_headers(url=request.url, total_count=0),
                data={
                    'method': request.method,
                    'url': url,
                    'url_root': url_root,
                    'path': path,
                    'query_string': query_string,
                    'resource_type': self.__class__.__name__,
                    'data': {'message': str(exception)},
                },
            )

        ## Validate request
        self.utils.validate_request(
            limit=limit,
            offset=offset,
            perpage=perpage,
            page=page,
            query_type=query_type,
            is_querystring_defined=(bool(query_string)),
        )

        ## Treat the projectable_properties case which does not imply access to the DataBase
        if query_type == 'projectable_properties':
            ## Retrieve the projectable properties
            projectable_properties, ordering = self.trans.get_projectable_properties()
            results = dict(fields=projectable_properties, ordering=ordering)
            ## Build response and return it
            headers = self.utils.build_headers(url=request.url, total_count=1)

        else:
            ## Set the query, and initialize qb object
            self.trans.set_query(filters=filters, orders=orderby, node_id=node_id)

            ## Count results
            total_count = self.trans.get_total_count()

            ## Pagination (if required)
            if page is not None:
                (limit, offset, rel_pages) = self.utils.paginate(page, perpage, total_count)
                self.trans.set_limit_offset(limit=limit, offset=offset)
                headers = self.utils.build_headers(rel_pages=rel_pages, url=request.url, total_count=total_count)
            else:
                self.trans.set_limit_offset(limit=limit, offset=offset)
                headers = self.utils.build_headers(url=request.url, total_count=total_count)

            ## Retrieve results
            results = self.trans.get_results()

        ## Build response and return it
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=request.path,
            id=node_id,
            query_string=query_string,
            resource_type=resource_type,
            data=results,
        )

        return self.utils.build_response(status=200, headers=headers, data=data)


class QueryBuilder(BaseResource):
    """Representation of a QueryBuilder REST API resource (instantiated with a serialised QueryBuilder instance).

    It supports POST requests taking in JSON :py:func:`~aiida.orm.querybuilder.QueryBuilder.as_dict`
    objects and returning the :py:class:`~aiida.orm.querybuilder.QueryBuilder` result accordingly.
    """

    _translator_class = NodeTranslator

    GET_MESSAGE = (
        'Method Not Allowed. Use HTTP POST requests to use the AiiDA QueryBuilder. '
        'POST JSON data, which MUST be a valid QueryBuilder.as_dict() dictionary as a JSON object. '
        'See the documentation at '
        'https://aiida.readthedocs.io/projects/aiida-core/en/latest/topics/database.html'
        '#converting-the-querybuilder-to-from-a-dictionary for more information.'
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # HTTP Request method decorators
        if 'get_decorators' in kwargs and isinstance(kwargs['get_decorators'], (tuple, list, set)):
            self.method_decorators.update({'post': list(kwargs['get_decorators'])})

    def get(self):
        """Static return to state information about this endpoint."""
        path, url, url_root, query_string = self.unquote_request()
        headers = self.utils.build_headers(url=request.url, total_count=1)
        return self.utils.build_response(
            status=405,  # Method Not Allowed
            headers=headers,
            data={
                'method': request.method,
                'url': url,
                'url_root': url_root,
                'path': path,
                'query_string': query_string,
                'resource_type': self.__class__.__name__,
                'data': {'message': self.GET_MESSAGE},
            },
        )

    def post(self):
        """POST method to pass query help JSON.

        If the posted JSON is not a valid QueryBuilder serialisation,
        the request will fail with an internal server error.

        This uses the NodeTranslator in order to best return Nodes according to the general AiiDA
        REST API data format, while still allowing the return of other AiiDA entities.

        :return: QueryBuilder result of AiiDA entities in "standard" REST API format.
        """
        path, url, url_root, query_string = self.unquote_request()
        profile = self.parse_query_string(query_string)[-1]

        try:
            self.load_profile(profile)
        except RestInputValidationError as exception:
            return self.utils.build_response(
                status=400,
                headers=self.utils.build_headers(url=request.url, total_count=0),
                data={
                    'method': request.method,
                    'url': url,
                    'url_root': url_root,
                    'path': path,
                    'query_string': query_string,
                    'resource_type': self.__class__.__name__,
                    'data': {'message': str(exception)},
                },
            )

        self.trans._query_help = request.get_json(force=True)
        # While the data may be correct JSON, it MUST be a single JSON Object,
        # equivalent of a QueryBuilder.as_dict() dictionary.
        assert isinstance(self.trans._query_help, dict), (
            'POSTed data MUST be a valid QueryBuilder.as_dict() dictionary. '
            f'Got instead (type: {type(self.trans._query_help)}): {self.trans._query_help}'
        )
        self.trans.__label__ = self.trans._result_type = self.trans._query_help['path'][-1]['tag']

        # Handle empty list projections
        number_projections = len(self.trans._query_help['project'])
        empty_projections_counter = 0
        skip_tags = []
        for tag, projections in tuple(self.trans._query_help['project'].items()):
            if projections == [{'*': {}}]:
                self.trans._query_help['project'][tag] = self.trans._default
            elif not projections:
                empty_projections_counter += 1
                skip_tags.append(tag)
            else:
                # Use projections as given, no need to "correct" them.
                pass

        if empty_projections_counter == number_projections:
            # No projections have been specified in the dictionary.
            # To be true to the QueryBuilder response, the last entry in path
            # is the only entry to be returned, all without edges/links.
            self.trans._query_help['project'][self.trans.__label__] = self.trans._default

        self.trans.init_qb()

        data = {}
        if self.trans.get_total_count():
            if empty_projections_counter == number_projections:
                # "Normal" REST API retrieval can be used.
                data = self.trans.get_results()
            else:
                # Since the "normal" REST API retrieval relies on single-tag retrieval,
                # we must instead be more creative with how we retrieve the results here.
                # So we opt for a dictionary, with the tags being the keys.
                for tag in self.trans._query_help['project']:
                    if tag in skip_tags:
                        continue
                    self.trans.__label__ = tag
                    data.update(self.trans.get_formatted_result(tag))

        # Remove 'full_type's when they're `None`
        for tag, entities in list(data.items()):
            updated_entities = []
            for entity in entities:
                if entity.get('full_type') is None:
                    entity.pop('full_type', None)
                updated_entities.append(entity)
            data[tag] = updated_entities

        headers = self.utils.build_headers(url=request.url, total_count=self.trans.get_total_count())
        return self.utils.build_response(
            status=200,
            headers=headers,
            data={
                'method': request.method,
                'url': unquote(request.url),
                'url_root': unquote(request.url_root),
                'path': unquote(request.path),
                'query_string': request.query_string.decode('utf-8'),
                'resource_type': self.__class__.__name__,
                'data': data,
            },
        )


class Node(BaseResource):
    """Differs from BaseResource in trans.set_query() mostly because it takes
    query_type as an input and the presence of additional result types like "tree"
    """

    _translator_class = NodeTranslator
    _parse_pk_uuid = 'uuid'  # Parse a uuid pattern in the URL path (not a pk)

    def get(self, id=None, page=None):
        """Get method for the Node resource.

        :param id: node identifier
        :param page: page no, used for pagination
        :return: http response
        """
        path, url, url_root, query_string = self.unquote_request()
        (resource_type, page, node_id, query_type) = self.parse_path(path)
        (
            limit,
            offset,
            perpage,
            orderby,
            filters,
            download_format,
            download,
            filename,
            tree_in_limit,
            tree_out_limit,
            attributes,
            attributes_filter,
            extras,
            extras_filter,
            full_type,
            profile,
        ) = self.parse_query_string(query_string)

        try:
            self.load_profile(profile)
        except RestInputValidationError as exception:
            return self.utils.build_response(
                status=400,
                headers=self.utils.build_headers(url=request.url, total_count=0),
                data={
                    'method': request.method,
                    'url': url,
                    'url_root': url_root,
                    'path': path,
                    'query_string': query_string,
                    'resource_type': self.__class__.__name__,
                    'data': {'message': str(exception)},
                },
            )

        ## Validate request
        self.utils.validate_request(
            limit=limit,
            offset=offset,
            perpage=perpage,
            page=page,
            query_type=query_type,
            is_querystring_defined=(bool(query_string)),
        )

        ## Treat the projectable properties case which does not imply access to the DataBase
        if query_type == 'projectable_properties':
            ## Retrieve the projectable properties
            projectable_properties, ordering = self.trans.get_projectable_properties()
            results = dict(fields=projectable_properties, ordering=ordering)
            ## Build response and return it
            headers = self.utils.build_headers(url=request.url, total_count=1)

        ## Treat the statistics
        elif query_type == 'statistics':
            headers = self.utils.build_headers(url=request.url, total_count=0)
            if filters:
                user_pk = filters['user']['==']
            else:
                user_pk = None
            results = self.trans.get_statistics(user_pk)

        elif query_type == 'full_types':
            headers = self.utils.build_headers(url=request.url, total_count=0)
            results = self.trans.get_namespace()

        elif query_type == 'full_types_count':
            headers = self.utils.build_headers(url=request.url, total_count=0)
            if filters:
                user_pk = filters['user']['==']
            else:
                user_pk = None
            results = self.trans.get_namespace(user_pk=user_pk, count_nodes=True)

        # TODO improve the performance of tree endpoint by getting the data from database faster
        # TODO add pagination for this endpoint (add default max limit)
        elif query_type == 'tree':
            headers = self.utils.build_headers(url=request.url, total_count=0)
            results = self.trans.get_io_tree(node_id, tree_in_limit, tree_out_limit)

        elif node_id is None and query_type == 'download_formats':
            headers = self.utils.build_headers(url=request.url, total_count=0)
            results = self.trans.get_all_download_formats(full_type)

        else:
            ## Initialize the translator
            self.trans.set_query(
                filters=filters,
                orders=orderby,
                query_type=query_type,
                node_id=node_id,
                download_format=download_format,
                download=download,
                filename=filename,
                attributes=attributes,
                attributes_filter=attributes_filter,
                extras=extras,
                extras_filter=extras_filter,
                full_type=full_type,
            )

            ## Count results
            total_count = self.trans.get_total_count()

            ## Pagination (if required)
            if page is not None:
                (limit, offset, rel_pages) = self.utils.paginate(page, perpage, total_count)
                self.trans.set_limit_offset(limit=limit, offset=offset)

                ## Retrieve results
                results = self.trans.get_results()

                headers = self.utils.build_headers(rel_pages=rel_pages, url=request.url, total_count=total_count)
            else:
                self.trans.set_limit_offset(limit=limit, offset=offset)
                ## Retrieve results
                results = self.trans.get_results()

                if query_type == 'repo_contents' and results:
                    response = make_response(results)
                    response.headers['content-type'] = 'application/octet-stream'
                    response.headers['Content-Disposition'] = f'attachment; filename="{filename}"'
                    return response

                if query_type == 'download' and download not in ['false', 'False', False] and results:
                    if results['download']['status'] == 200:
                        data = results['download']['data']
                        response = make_response(data)
                        response.headers['content-type'] = 'application/octet-stream'
                        response.headers['Content-Disposition'] = 'attachment; filename="{}"'.format(
                            results['download']['filename']
                        )
                        return response

                    results = results['download']['data']

                headers = self.utils.build_headers(url=request.url, total_count=total_count)

            if attributes_filter is not None and attributes:
                for node in results['nodes']:
                    node['attributes'] = {}
                    if not isinstance(attributes_filter, list):
                        attributes_filter = [attributes_filter]
                    for attr in attributes_filter:
                        node['attributes'][str(attr)] = node[f'attributes.{attr!s}']
                        del node[f'attributes.{attr!s}']

            if extras_filter is not None and extras:
                for node in results['nodes']:
                    node['extras'] = {}
                    if not isinstance(extras_filter, list):
                        extras_filter = [extras_filter]
                    for extra in extras_filter:
                        node['extras'][str(extra)] = node[f'extras.{extra!s}']
                        del node[f'extras.{extra!s}']

        ## Build response
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=path,
            id=node_id,
            query_string=query_string,
            resource_type=resource_type,
            data=results,
        )

        return self.utils.build_response(status=200, headers=headers, data=data)


class Computer(BaseResource):
    """Resource for Computer"""

    from aiida.restapi.translator.computer import ComputerTranslator

    _translator_class = ComputerTranslator
    _parse_pk_uuid = 'uuid'


class Group(BaseResource):
    """Resource for Group"""

    from aiida.restapi.translator.group import GroupTranslator

    _translator_class = GroupTranslator
    _parse_pk_uuid = 'uuid'


class User(BaseResource):
    """Resource for User"""

    from aiida.restapi.translator.user import UserTranslator

    _translator_class = UserTranslator
    _parse_pk_uuid = 'pk'


class ProcessNode(Node):
    """Resource for ProcessNode"""

    from aiida.restapi.translator.nodes.process.process import ProcessTranslator

    _translator_class = ProcessTranslator

    def get(self, id=None, page=None):
        """Get method for the Process resource.

        :param id: node identifier
        :return: http response
        """
        path, url, url_root, query_string = self.unquote_request()
        resource_type, page, node_id, query_type = self.parse_path(path)
        profile = self.parse_query_string(query_string)[-1]

        try:
            self.load_profile(profile)
        except RestInputValidationError as exception:
            return self.utils.build_response(
                status=400,
                headers=self.utils.build_headers(url=request.url, total_count=0),
                data={
                    'method': request.method,
                    'url': url,
                    'url_root': url_root,
                    'path': path,
                    'query_string': query_string,
                    'resource_type': self.__class__.__name__,
                    'data': {'message': str(exception)},
                },
            )

        headers = self.utils.build_headers(url=request.url, total_count=1)

        results = None
        if query_type == 'report':
            node = self._load_and_verify(node_id)
            report = self.trans.get_report(node)
            results = report

        elif query_type == 'projectable_properties':
            ## Retrieve the projectable properties
            projectable_properties, ordering = self.trans.get_projectable_properties()
            results = dict(fields=projectable_properties, ordering=ordering)

        ## Build response
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=path,
            id=node_id,
            query_string=query_string,
            resource_type=resource_type,
            data=results,
        )

        return self.utils.build_response(status=200, headers=headers, data=data)


class CalcJobNode(ProcessNode):
    """Resource for CalcJobNode"""

    from aiida.restapi.translator.nodes.process.calculation.calcjob import CalcJobTranslator

    _translator_class = CalcJobTranslator

    def get(self, id=None, page=None):
        """Get method for the Process resource.

        :param id: node identifier
        :return: http response
        """
        path, url, url_root, query_string = self.unquote_request()
        resource_type, page, node_id, query_type = self.parse_path(path)
        profile = self.parse_query_string(query_string)[-1]

        try:
            self.load_profile(profile)
        except RestInputValidationError as exception:
            return self.utils.build_response(
                status=400,
                headers=self.utils.build_headers(url=request.url, total_count=0),
                data={
                    'method': request.method,
                    'url': url,
                    'url_root': url_root,
                    'path': path,
                    'query_string': query_string,
                    'resource_type': self.__class__.__name__,
                    'data': {'message': str(exception)},
                },
            )

        node = self._load_and_verify(node_id)
        results = None

        params = request.args
        filename = params.get('filename', '')

        if query_type == 'input_files':
            results = self.trans.get_input_files(node, filename)
        elif query_type == 'output_files':
            results = self.trans.get_output_files(node, filename)

        ## Build response and return it
        headers = self.utils.build_headers(url=request.url, total_count=1)

        ## Build response
        data = dict(
            method=request.method,
            url=url,
            url_root=url_root,
            path=path,
            id=node_id,
            query_string=query_string,
            resource_type=resource_type,
            data=results,
        )

        return self.utils.build_response(status=200, headers=headers, data=data)
